searchData={"items":[{"type":"task","doc":"Assembles an Altworx scenario release for the current project. It is done in\na docker container based on an image which ensures correct erlang and Elixir\nversions, so they are compatible with Altworx.\n\n    $ mix altworx.scenario_release","title":"mix altworx.scenario_release","ref":"Mix.Tasks.Altworx.ScenarioRelease.html"},{"type":"module","doc":"Macro for emulating a StageBased scenario inside a Simple scenario.\n\nThis macro provides a simple way to integrate `Runbox.Runtime.Stage.Emulator` in a Simple\nscenario.\n\nThis macro has to be used in a manifest module of a scenario. The manifest must specify the type\nof the scenario as `Runbox.Scenario.Type.simple/0`. The rest is taken care of by the macro.\n\nIt does the following\n- use `Runbox.Scenario.Simple` to register the module as the implementation of the scenario\n- implement all necessary callbacks and use `Runbox.Runtime.Stage.Emulator` in them\n- auto-discover template modules, see below","title":"Runbox.Runtime.Simple.StageBasedEmulator","ref":"Runbox.Runtime.Simple.StageBasedEmulator.html"},{"type":"module","doc":"This macro auto-discovers template modules of the emulated scenario. This is designed to easily\nswitch a StageBased scenario into a Simple scenario emulating the original code.\n\nTemplate modules are discovered as follows:\n- list all modules in the scenario release\n- take only modules whose name are prefixed by the current module, just like a StageBased\n  templates are discovered\n- from these take only modules which are valid templates, basically this means it has to use the\n  behavior `Runbox.Scenario.Template.StageBased`.\n\nTherefore, to convert existing StageBased scenario to Simple scenario you only have to do the\nfollowing:\n- add `use Runbox.Runtime.Simple.StageBasedEmulator` to the manifest module\n- change type in the manifest to `Runbox.Scenario.Type.simple/0`","title":"Template module auto-discovery - Runbox.Runtime.Simple.StageBasedEmulator","ref":"Runbox.Runtime.Simple.StageBasedEmulator.html#module-template-module-auto-discovery"},{"type":"module","doc":"```elixir\ndefmodule Scenarios.MyScenario do\n  use Runbox.Scenario.Manifest\n  use Runbox.Runtime.Simple.StageBasedEmulator\n\n  @impl Runbox.Scenario.Manifest\n  def get_info do\n    %Runbox.Scenario.Manifest{\n      id: \"my_scenario\",\n      name: \"My converted scenario\",\n      description: \"This is a scenario I converted from StageBased to Simple\",\n      change_log: %{\n        1704021223000 => \"First version of the scenario\",\n        1731487296000 => \"Convert to Simple scenario\"\n      },\n      type: Runbox.Scenario.Type.simple()\n    }\n  end\nend\n```","title":"Example - Runbox.Runtime.Simple.StageBasedEmulator","ref":"Runbox.Runtime.Simple.StageBasedEmulator.html#module-example"},{"type":"module","doc":"StageBased scenario runtime single-process emulator.\n\nUsing this module you can emulate multi-process StageBased scenario inside a single process. The\nemulation uses the original templates, i.e. modules implementing\n`Runbox.Scenario.Template.StageBased` behavior.\n\nThe emulation is designed to emulate a StageBased scenario in another scenario, usually in another\nscenario type (like Simple). You can also use this to test StageBased scenarios, however,\n`Runbox.Runtime.Stage.Sandbox` might be a better fit for testing.\n\nIf you plan to use this in a Simple scenario checkout a macro\n`Runbox.Runtime.Simple.StageBasedEmulator`. It provides deeper integration with Simple scenario\nruntime and enables you to quickly and easily convert StageBased scenario to Simple scenario\nemulating the original behavior.","title":"Runbox.Runtime.Stage.Emulator","ref":"Runbox.Runtime.Stage.Emulator.html"},{"type":"module","doc":"The emulation consists of the following steps and activities. The emulated scenario is defined by\na list of modules - templates. The emulation has its own internal state, so you don't need to\nremember the modules yourself.\n\nFirst it's usually needed to know which logical topics are required by the emulated scenario. You\ncan use function `input_topics/1` for this purpose. It returns list of all input topics defined in\ntemplates' subscriptions as well as the type of the topic (input or load).\n\nThen you can actually start executing the emulated scenario. At first you need to initialize it\nvia `initialize/2`. This initializes the internals of the emulator, creates all static units\n(those defined in `c:Runbox.Scenario.Template.StageBased.instances/0`) in every template and also\nexecutes the `c:Runbox.Scenario.Template.StageBased.init/2` callback for each of these units.\n\nThen you can start handling messages via the function `handle_message/3`. Apart from regular\nscenario outputs, like output actions, this also outputs timeout registrations that need to be\nhandled externally. This is done because handling timeouts internally might lead to problems when\nusing the emulator in a scenario.\n\nIf you are persisting the internal state, such as in a savepoint, you need to use\n`reinitialize_state/1` to refresh information that might not survive a persistence. Note the state\nis only designed to be persisted via ETF, it doesn't support JSON or other more restrictive\nformats.","title":"Example use - Runbox.Runtime.Stage.Emulator","ref":"Runbox.Runtime.Stage.Emulator.html#module-example-use"},{"type":"function","doc":"Handles a message in the emulation session.\n\nTakes a message, the logical topic it came from and the state of the emulation session and handles\nthe message by the emulated scenario. This can produce outputs and also can change the state of\nthe emulated scenario.\n\nOne of the outputs is a `Runbox.Runtime.RuntimeInstruction`. The caller is responsible for\nevaluating these instructions. E.g. if a timeout is registered the caller must ensure the timeout\nmessage is handled at the appropriate time.\n\nNote that when handling such timeouts the argument `from` is ignored. Timeout messages are\nautomatically routed to the unit they were created in and are not subjected to the regular message\nrouting.","title":"Runbox.Runtime.Stage.Emulator.handle_message/3","ref":"Runbox.Runtime.Stage.Emulator.html#handle_message/3"},{"type":"function","doc":"Initializes an emulation session.\n\nThe emulation is started by an initialization. This takes a list of templates and parameters\nneeded for their initialization. The templates are used to create an emulated version of the\ncomputation network, much like a true StageBased scenario. This is encapsulated in an internal\nstate. The state holds all the information needed for the emulation, including the current scenario\nstate, and is then passed to `handle_message/3` when consuming messages.\n\nStatically defined units (defined via `c:Runbox.Scenario.Template.StageBased.instances/0`) are\ninstantiated and initialized. This can produce output actions which are returned from this\nfunction.\n\nOne of the outputs is a `Runbox.Runtime.RuntimeInstruction`. The caller is responsible for\nevaluating these instructions. E.g. if a timeout is registered the caller must ensure the timeout\nmessage is handled at the appropriate time.","title":"Runbox.Runtime.Stage.Emulator.initialize/2","ref":"Runbox.Runtime.Stage.Emulator.html#initialize/2"},{"type":"function","doc":"Returns input topics the templates are subscribed to.\n\nGoes through every specified template and finds all input topics in the subscriptions. Returns a\nlist of all topics the templates are subscribed to including the type of the topic.","title":"Runbox.Runtime.Stage.Emulator.input_topics/1","ref":"Runbox.Runtime.Stage.Emulator.html#input_topics/1"},{"type":"function","doc":"Re-initializes a state after it has been persisted and loaded.\n\nThe emulation state can be persisted into ETF. However, when such state is loaded some internal\nreferences might not work correctly and it is required to re-initialize the state. Call this\nfunction after you load the state and before you start using it.","title":"Runbox.Runtime.Stage.Emulator.reinitialize_state/1","ref":"Runbox.Runtime.Stage.Emulator.html#reinitialize_state/1"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.Emulator.output/0","ref":"Runbox.Runtime.Stage.Emulator.html#t:output/0"},{"type":"opaque","doc":"","title":"Runbox.Runtime.Stage.Emulator.state/0","ref":"Runbox.Runtime.Stage.Emulator.html#t:state/0"},{"type":"opaque","doc":"","title":"Runbox.Runtime.Stage.Emulator.timeout_message_type/0","ref":"Runbox.Runtime.Stage.Emulator.html#t:timeout_message_type/0"},{"type":"module","doc":"Struct representing a unit.","title":"Runbox.Runtime.Stage.Unit","ref":"Runbox.Runtime.Stage.Unit.html"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.Unit.attributes/1","ref":"Runbox.Runtime.Stage.Unit.html#attributes/1"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.Unit.id/1","ref":"Runbox.Runtime.Stage.Unit.html#id/1"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.Unit.new/3","ref":"Runbox.Runtime.Stage.Unit.html#new/3"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.Unit.set_state/2","ref":"Runbox.Runtime.Stage.Unit.html#set_state/2"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.Unit.state/1","ref":"Runbox.Runtime.Stage.Unit.html#state/1"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.Unit.t/0","ref":"Runbox.Runtime.Stage.Unit.html#t:t/0"},{"type":"module","doc":"Provides configuration to a scenario.","title":"Runbox.Scenario.Config","ref":"Runbox.Scenario.Config.html"},{"type":"function","doc":"Read file from the config dir.\n\nThe file is located in the configuration directory, in a folder matching the `scenario_id` param.\nThe file is read if it exists and content is returned as `{:ok, content}`, otherwise `{:error,\nreason}` is returned.","title":"Runbox.Scenario.Config.read_file/2","ref":"Runbox.Scenario.Config.html#read_file/2"},{"type":"function","doc":"Resolves a path to a file in scenario config.\n\nThe file is located in the configuration directory, in a folder matching the `scenario_id` param.\nThis function resolves the path so that it can be used with `File` functions directly.  The\nexistence of the file is not verified. Note if you want to just read the file, you can use\n`read_file/2`, this function should be used when you need to handle the file in a special manner\n(e.g. streaming).","title":"Runbox.Scenario.Config.resolve_file/2","ref":"Runbox.Scenario.Config.html#resolve_file/2"},{"type":"module","doc":"Support for helper processes for scenarios.\n\nProvides a supervisor and process registry under which scenarios can start\ntheir side-kick processes intended for housekeeping tasks.","title":"Runbox.Scenario.Helper","ref":"Runbox.Scenario.Helper.html"},{"type":"function","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","title":"Runbox.Scenario.Helper.child_spec/1","ref":"Runbox.Scenario.Helper.html#child_spec/1"},{"type":"function","doc":"Starts a supervisor with dynamic supervisor and process registry.","title":"Runbox.Scenario.Helper.start_link/1","ref":"Runbox.Scenario.Helper.html#start_link/1"},{"type":"function","doc":"Returns a `:via` name to be used in the helper process registry.","title":"Runbox.Scenario.Helper.via_registry/1","ref":"Runbox.Scenario.Helper.html#via_registry/1"},{"type":"function","doc":"Executes a given `callback` over running helper.\n\nChecks whether the helper with `name` registered under helper registry is\nalready running. If not, then it starts the process according to `child_spec`\nunder the dynamic supervisor. In either case, it passes the helper process\npid to the `callback`.\n\nIt is the responsibility of the helper process to register its name with the\nhelper registry via the `via_registry/1`.","title":"Runbox.Scenario.Helper.with_helper/3","ref":"Runbox.Scenario.Helper.html#with_helper/3"},{"type":"function","doc":"iex> child_spec = %{\n    ...>   id: MyHelper,\n    ...>   start: {\n    ...>     Agent,\n    ...>     :start_link,\n    ...>     [fn -> 1 end, [name: Helper.via_registry(MyHelper)]]\n    ...>   }\n    ...> }\n    iex> :ok = Helper.with_helper(MyHelper, child_spec, fn pid -> Agent.update(pid, & &1 + 1) end)\n    iex> Helper.with_helper(MyHelper, child_spec, fn pid -> Agent.get(pid, & &1) end)\n    2","title":"Examples - Runbox.Scenario.Helper.with_helper/3","ref":"Runbox.Scenario.Helper.html#with_helper/3-examples"},{"type":"behaviour","doc":"Behaviour and struct for scenario manifest.\n\nManifest is the main module of a scenario, each scenario must have a single manifest module and\neach manifest defines a scenario. Manifest module provides basic information/metadata about a\nscenario. This is returned in a callback `c:get_info/0` which returns the struct defined in this\nmodule.\n\nThe struct contains the following information about a scenario:\n  * `id` - scenario identification.\n  * `name` - name of the scenario, should be short.\n  * `description` - more detailed description of the scenario, should be longer than name and more\n    descriptive\n  * `version` - scenario version, major versions are not compatible, so if you are upgrading to\n     scenario with different major version all runs of the scenario will be stopped and\n     must be recomputed manually.\n  * `change_log` - information about changes made to the scenario, for users displayed e.g. in\n    Admin UI.\n  * `type` - type of the scenario, see `Runbox.Scenario.Type` for supported types\n  * `incidents` - incidents created by this scenario and some metadata for each incident type. See\n    `t:Incident.t/0` for details.\n  * `events` - events created by this scenario and some metadata for each event type. \n  * `groups` - a list of groups this scenario can use and some metadata for each. See\n    `t:group_of_assets/0` for details.\n  * `required_raw_topics` - list of raw topics required by the scenario. These will be created\n    automatically when the scenario is loaded.\n  * `notifications` - metadata about notifications generated by this scenario. See\n    `t:notification_metadata/0` for details.\n\nSee `t:t/0` for type details.","title":"Runbox.Scenario.Manifest","ref":"Runbox.Scenario.Manifest.html"},{"type":"callback","doc":"Invoked to obtain the scenario manifest.","title":"Runbox.Scenario.Manifest.get_info/0","ref":"Runbox.Scenario.Manifest.html#c:get_info/0"},{"type":"callback","doc":"Invoked once when the scenario run starts.\n\nUseful for doing any necessary preparations for scenario run.\n\n**Warning:** Do not rely on any external resources (e.g. SQL queries) which may\nchange the scenario logic.\n\nThis callback is optional. If it is not implemented, no action is performed\nwhen starting the scenario run.","title":"Runbox.Scenario.Manifest.on_start/0","ref":"Runbox.Scenario.Manifest.html#c:on_start/0"},{"type":"type","doc":"Assets group attributes metadata.\n\nMetadata helps the UI to render form for the group attributes definition.\nAttributes of the group are saved in asset attributes in Reality Network\nif group is materialized as asset in Reality Network.","title":"Runbox.Scenario.Manifest.attribute_detail/0","ref":"Runbox.Scenario.Manifest.html#t:attribute_detail/0"},{"type":"type","doc":"* `:name` - user-friendly attribute name - what is set.\n  * `:description` - longer attribute description - what the attribute does, the unit, ...\n  * `:path` - path to attribute in attributes JSON.\n  * `:type` - type of attribute value (helps the UI to render form input).","title":"Properties - Runbox.Scenario.Manifest.attribute_detail/0","ref":"Runbox.Scenario.Manifest.html#t:attribute_detail/0-properties"},{"type":"type","doc":"Metadata (template) of assets group.\n\nThe scenario should be able to carry a list of groups it uses. These groups\ncan then be easily created in the UI by the user without having to fill in\nsome information manually. Metadata defines these helpers for the UI.","title":"Runbox.Scenario.Manifest.group_of_assets/0","ref":"Runbox.Scenario.Manifest.html#t:group_of_assets/0"},{"type":"type","doc":"* `:name` - short name of group template for UI.\n  * `:description` - description of assets group.\n  * `:asset_type` - asset type of asset created by the group.\n  * `:edge_type` - edge type for connecting assets in the group.\n  * `:edge_direction` - edge direction (true = from group asset to assets in group).\n  * `:attributes` - description of group attributes.","title":"Properties - Runbox.Scenario.Manifest.group_of_assets/0","ref":"Runbox.Scenario.Manifest.html#t:group_of_assets/0-properties"},{"type":"type","doc":"Notification specification metadata.","title":"Runbox.Scenario.Manifest.notification_metadata/0","ref":"Runbox.Scenario.Manifest.html#t:notification_metadata/0"},{"type":"type","doc":"* `:name` - user-friendly name for the notification type\n  * `:metadata_types` - describes metadata the notification caries, this information can be used for\n    additional filtering in subscriptions. It can be empty map if no metadata is needed.\n  * `type_name` - is a binary specifying the name of the notification metadata type\n  * `:type` - type of notification metadata, can be either a `:number`, `:string` or `:boolean`.","title":"Properties - Runbox.Scenario.Manifest.notification_metadata/0","ref":"Runbox.Scenario.Manifest.html#t:notification_metadata/0-properties"},{"type":"type","doc":"Notification type is an atom uniquely identifying the notification in the scenario","title":"Runbox.Scenario.Manifest.notification_type/0","ref":"Runbox.Scenario.Manifest.html#t:notification_type/0"},{"type":"type","doc":"The struct contains information about a scenario.\n\nSee [module documentation](#moduledoc) for detailed description of the fields.","title":"Runbox.Scenario.Manifest.t/0","ref":"Runbox.Scenario.Manifest.html#t:t/0"},{"type":"module","doc":"Struct representing incident manifest","title":"Runbox.Scenario.Manifest.Incident","ref":"Runbox.Scenario.Manifest.Incident.html"},{"type":"type","doc":"","title":"Runbox.Scenario.Manifest.Incident.t/0","ref":"Runbox.Scenario.Manifest.Incident.html#t:t/0"},{"type":"module","doc":"Struct representing one column in incident forecast UI widget.","title":"Runbox.Scenario.Manifest.IncidentFrame","ref":"Runbox.Scenario.Manifest.IncidentFrame.html"},{"type":"module","doc":"* `:name` displayed column name\n  * `:since` time frame start interval in ms\n  * `:till` time frame end interval in ms","title":"Properties: - Runbox.Scenario.Manifest.IncidentFrame","ref":"Runbox.Scenario.Manifest.IncidentFrame.html#module-properties"},{"type":"type","doc":"","title":"Runbox.Scenario.Manifest.IncidentFrame.t/0","ref":"Runbox.Scenario.Manifest.IncidentFrame.html#t:t/0"},{"type":"behaviour","doc":"Simple scenario behavior.\n\nImplement this behaviour to create a scenario. Note the scenario must also implement\n`Runbox.Scenario.Manifest`, although all this can be done in a single module, see example below.","title":"Runbox.Scenario.Simple","ref":"Runbox.Scenario.Simple.html"},{"type":"behaviour","doc":"Simple scenario has a single module that carries the whole logic of the scenario. The module\nspecifies which logical topics it wants to consume (`c:input_topics/0`) and defines how messages from\nthese topics are processed, updating its state and generating outputs along the way.\n\nUsually you would use a single module for a scenario implementing both Manifest and this behavior.\nHowever, you can separate these two if you wish. Note that the name of the module must always be\nprefixed by the module name structure of the Manifest. E.g. if you have manifest module\n`Scenarios.MyScenario` the module implementing this behaviour can be named\n`Scenarios.MyScenario.Logic` but not `Scenarios.MyScenarioLogic`.\n\nBefore the processing is first started an `c:init/1` callback is used to initialize the state.\nState is later used in the computation.\n\nEach message in the specified topics is then processed by the `c:handle_message/2` callback. This\ncallback should contain the main logic of the scenario. Messages arrive sorted into this callback -\nyou can assume the time is non-decreasing.\n\nAt minimum you must implement callbacks  `c:input_topics/0` and `c:handle_message/2`, see other\ncallbacks available in this module and in `Runbox.Scenario.Manifest` to see what is possible in a\nscenario.","title":"Simple scenario structure - Runbox.Scenario.Simple","ref":"Runbox.Scenario.Simple.html#module-simple-scenario-structure"},{"type":"behaviour","doc":"Scenarios should be deterministic, meaning that if you run a scenario twice on the same data you\nget the same results. Scenarios shouldn't depend on any external resources other than topics and\nscenario configuration. Notably, scenarios shouldn't depend on current time. Scenario should\nalways only work with time from the incoming messages, called *external time*.\n\nYou should avoid directly making side effects in a scenario. Scenario should instead produce\n*output actions*, which are instructions to perform specific side effects. Output actions are\nautomatically performed by the scenario runtime in a deterministic manner. See\n`Runbox.Scenario.OutputAction` for the list of available output actions.\n\nThere are cases when you want to do some work at a specified point of the external time, for\nexample to produce some output actions when the external time reaches noon. This can be also\nachieved in a deterministic manner via *timeouts*. You can register a timeout by returning\na corresponding *runtime instruction*, see `Runbox.Runtime.RuntimeInstruction.register_timeout/1`\nfor more details. The message registered in a timeout will be then passed as is to your\n`c:handle_message/2` callback exactly at the time you specify in its `timestamp` field.","title":"Determinism - Runbox.Scenario.Simple","ref":"Runbox.Scenario.Simple.html#module-determinism"},{"type":"behaviour","doc":"State used in the computation is automatically and periodically persisted. This makes it easier to\nrestart the computation. You don't have to recompute all of the data again to get the final state,\nyou just load the last savepoint and recompute from there.\n\nBy default the state is persisted in the same structure as is used in the scenario. However, if\nyou wish you can implement `c:get_state/1` and `c:set_state/1` callbacks to transform the state\nbefore it is saved into a savepoint and after it is loaded from a savepoint.","title":"State persistence - Runbox.Scenario.Simple","ref":"Runbox.Scenario.Simple.html#module-state-persistence"},{"type":"behaviour","doc":"defmodule Scenarios.ServerTemperature do\n      use Runbox.Scenario.Manifest\n      use Runbox.Scenario.Simple\n\n      alias Runbox.Scenario.Manifest\n      alias Runbox.Scenario.OutputAction.Event\n      alias Runbox.Scenario.OutputAction.UpsertAssetAttributes\n      alias Runbox.Scenario.Type, as: ScenarioType\n      alias Runbox.Message\n\n      @impl Manifest\n      def get_info do\n        %Manifest{\n          id: \"server_temperature\",\n          name: \"Server temperature\",\n          description: \"Monitors server temperature and saves the data into Reality Network.\",\n          type: ScenarioType.simple()\n        }\n      end\n\n      @impl Runbox.Scenario.Simple\n      def init(_) do\n        {:ok, [], %{last_temps: %{}}}\n      end\n\n      @impl Runbox.Scenario.Simple\n      def input_topics do\n        [\"server_sensor\"]\n      end\n\n      @impl Runbox.Scenario.Simple\n      def handle_message(%Message{type: :temperature} = msg, state) do\n        last_temp = Map.get(state.last_temps, msg.body.server, 0)\n\n        if last_temp != msg.body.temperature do\n          # temperature changed, update reality network\n          event = %Event{\n            type: \"temp_change\",\n            template: \"Temperature on ${actors.server} changed to ${params.temp}\",\n            actors: %{\"server\" => %{asset_type: \"/assets/servers\", asset_id: msg.body.server}},\n            params: %{\"temp\" => msg.body.temperature},\n            origin_messages: [msg.origin]\n          }\n\n          update_asset = %UpsertAssetAttributes{\n            type: \"/assets/servers\",\n            id: msg.body.server,\n            attributes: %{\"temp\" => msg.body.temperature}\n          }\n\n          new_state = put_in(state.last_temps[msg.body.server], msg.body.temperature)\n          {:ok, [event, update_asset], new_state}\n        else\n          # temperature didn't change, no need to do anything\n          {:ok, [], state}\n        end\n      end\n\n      # ignore other sensor data\n      def handle_message(_, state) do\n        {:ok, [], state}\n      end\n    end","title":"Example - Runbox.Scenario.Simple","ref":"Runbox.Scenario.Simple.html#module-example"},{"type":"callback","doc":"Convert internal state to external.\n\nCalled before a state is persisted, the callback should convert the internal state of the to the\nexternal state, which is ready for persistence. If the scenario uses any other resources that are\nnot directly stored in the internal state, they should be stored in the external state in this\nstep.\n\nThis callback is optional, defaulting to `Function.identity/1`. Therefore, you should implement this\ncallback only when the external state differs from the internal.","title":"Runbox.Scenario.Simple.get_state/1","ref":"Runbox.Scenario.Simple.html#c:get_state/1"},{"type":"callback","doc":"Handle a message from input topics.\n\nFor each message in the input topics this callback is called. This is where most of the scenario\nbusiness logic lives.\n\nThe computation shouldn't perform any direct side effects. Instead it should rely on output\nactions and runtime instructions that are returned from the callback. Output actions are performed\nat the time of the currently handled message.\n\nSince scenario is a stateful processing, you can use the internal state in the computation. You\nget the previous state as an argument and you can modify the state.\n\nThis callback is mandatory, without it the scenario has no meaning.","title":"Runbox.Scenario.Simple.handle_message/2","ref":"Runbox.Scenario.Simple.html#c:handle_message/2"},{"type":"callback","doc":"Initializes the scenario run.\n\nWhen a run is first started, this callback is called to initialize the state. Additionally you can\nalso generate some output actions or runtime instructions. Output actions are produced in the time\nof the `start_from` parameter (see `Runbox.Scenario.Simple.Config`).\n\nThe callback is optional, if not provided the state is initialized to `nil`.","title":"Runbox.Scenario.Simple.init/1","ref":"Runbox.Scenario.Simple.html#c:init/1"},{"type":"callback","doc":"Returns a list of topics this scenario consumes.\n\nEach scenario consumes one or more topics specified in this callback.\n\nTopic names in this callback are specified as *logical* topics. That is the name without\nAltworx-specific prefixes, e.g. `DATA_1`. It represents the logical data without any tie to the\nphysical representation. When a run is started physical topics are assigned to each logical topic.\nThat is the names of actual topics physically available on the Altworx instance. By default the\nAltworx-specific prefix is added to the logical name to form a physical topic name, e.g.\n`N6_rt_DATA_1`, but user can override this and choose any other physical topic name.\n\nConsumed topics can be of two different types:\n* input topic - the default type, regular input, data is read from where run was started\n* load topic - data is always read from the start irrespective of where run is started from. This\n  is useful for sparse configuration topics, or generally topics from which you always need the\n  complete information before starting processing other data.\n\nA single scenario can subscribe to multiple topics of various types.\n\nThis callback is mandatory.","title":"Runbox.Scenario.Simple.input_topics/0","ref":"Runbox.Scenario.Simple.html#c:input_topics/0"},{"type":"callback","doc":"Convert external state to internal and bootstrap the scenario run.\n\nCalled when loading a savepoint to start a run.\n\nThis callback converts the external persisted state into the internal state used for computation.\nThe callback can also be used to start any other resources it uses and bootstrap them with\ninformation from external state.\n\nThis callback is optional, defaulting to `Function.identity/1`. Therefore, you should implement\nthis callback only when the external state differs from the internal or when you need to bootstrap\nsome other resources.","title":"Runbox.Scenario.Simple.set_state/1","ref":"Runbox.Scenario.Simple.html#c:set_state/1"},{"type":"type","doc":"External state of the scenario used for persistence.\n\nWhen scenario needs to save its state into a savepoint, it converts the internal state into an\nexternal state. For this purpose a callback `c:get_state/1` is used. Similarly, when a savepoint\nis loaded the external state inside is converted into the internal state, which is given to the\nscenario so it can start processing messages. A callback `c:set_state/1` is used for that purpose.","title":"Runbox.Scenario.Simple.external_state/0","ref":"Runbox.Scenario.Simple.html#t:external_state/0"},{"type":"type","doc":"Scenario outputs.\n\nAn output can be either an output actions or a runtime instructions.","title":"Runbox.Scenario.Simple.outputs/0","ref":"Runbox.Scenario.Simple.html#t:outputs/0"},{"type":"type","doc":"Internal state of the scenario used for computation.\n\nThis is what the scenario uses for its computation. This state is bootstrapped in `c:init/1` and\nis then used and modified in `c:handle_message/2`.","title":"Runbox.Scenario.Simple.state/0","ref":"Runbox.Scenario.Simple.html#t:state/0"},{"type":"module","doc":"Simple scenario configuration struct.\n\nThis struct is passed to the scenario when it is started. It contains the following fields:\n\n*  `start_from` - integer, ms from epoch, from when is run started to read data","title":"Runbox.Scenario.Simple.Config","ref":"Runbox.Scenario.Simple.Config.html"},{"type":"type","doc":"Scenario configuration passed to the scenario when it is started.\n\nSee moduledoc for more information.","title":"Runbox.Scenario.Simple.Config.t/0","ref":"Runbox.Scenario.Simple.Config.html#t:t/0"},{"type":"behaviour","doc":"Module defines behaviour used in scenarios as template.\n\nScenario template implementations are main components of scenario. Runtime uses callbacks of\nthis behaviour to start scenario in run and to handle incoming messages. Each template\nbehaviour implementation is represented as separate process in runtime called `TemplateCarrier`.\n\nTemplate module name structure must be always prefixed by the module name structure of the\nManifest. E.g. if you have manifest module `Scenarios.MyScenario` a template can be named\n`Scenarios.MyScenario.Templates.Car` or just `Scenarios.MyScenario.Car` but not\n`Scenarios.MyScenarioCar`.\n\nTemplate can also be defined in the same module as the Manifest. This is useful in cases where\nthere is only a single template in the scenario.\n\nTemplate is intended to be used to model behaviour of assets of single type per template.\nIndividual asset states are represented as `t:Runbox.Runtime.Stage.Unit.t/0`. Initial assets are\ndefined via `c:Runbox.Scenario.Template.StageBased.instances/0` callback. This callback should\nreturn list of unit definitions (`t:Runbox.Scenario.Template.StageBased.instances_def/0`).\n\nTemplates are linked together via subscriptions provided by callback\n`c:Runbox.Scenario.Template.StageBased.subscriptions/0`. Subscription is defined as tuple\ncontaining component definition and routing rule list. It is possible to subscribe to other\ntemplates, input and load topics. Routing rules are used to locate units when message is handled.\nWhen routing rule succeeds, callback `c:Runbox.Scenario.Template.StageBased.handle_message/2`\nis called, when routing rule fails to locat existing unit,\n`c:Runbox.Scenario.Template.StageBased.handle_asset_discovery/1` gets called instead.\n\nUnit initialization can produce side effects via `c:Runbox.Scenario.Template.StageBased.init/2`\ncallback. This callback is triggered when instances are produced by\n`c:Runbox.Scenario.Template.StageBased.instances/0` or\n`c:Runbox.Scenario.Template.StageBased.handle_asset_discovery/1` callbacks. Units are already\nregistered in the time when this callback is triggered, so changes in unit attributes won't\nchange unit registration (defined by `c:Runbox.Scenario.Template.StageBased.subscriptions/0`).\n\nWhen message from subscribed publisher is handled by the `TemplateCarrier`, runtime tries to\nlocate asset unit using given subscription configuration. If there is such a unit, runtime\ncalls `c:Runbox.Scenario.Template.StageBased.handle_message/2` callback and uses message and\nlocated message as attributes. In case that no unit was found,\n`c:Runbox.Scenario.Template.StageBased.handle_asset_discovery/1` gets called instead.","title":"Runbox.Scenario.Template.StageBased","ref":"Runbox.Scenario.Template.StageBased.html"},{"type":"callback","doc":"","title":"Runbox.Scenario.Template.StageBased.handle_asset_discovery/1","ref":"Runbox.Scenario.Template.StageBased.html#c:handle_asset_discovery/1"},{"type":"function","doc":"Invokes `template_module.asset_discovery` with given `message`.","title":"Runbox.Scenario.Template.StageBased.handle_asset_discovery/2","ref":"Runbox.Scenario.Template.StageBased.html#handle_asset_discovery/2"},{"type":"callback","doc":"","title":"Runbox.Scenario.Template.StageBased.handle_message/2","ref":"Runbox.Scenario.Template.StageBased.html#c:handle_message/2"},{"type":"function","doc":"Invokes `template_module.handle_message` with given `message` and `unit`.","title":"Runbox.Scenario.Template.StageBased.handle_message/3","ref":"Runbox.Scenario.Template.StageBased.html#handle_message/3"},{"type":"callback","doc":"","title":"Runbox.Scenario.Template.StageBased.init/2","ref":"Runbox.Scenario.Template.StageBased.html#c:init/2"},{"type":"function","doc":"Updates unit and sets initial unit state.","title":"Runbox.Scenario.Template.StageBased.init/3","ref":"Runbox.Scenario.Template.StageBased.html#init/3"},{"type":"callback","doc":"","title":"Runbox.Scenario.Template.StageBased.instances/0","ref":"Runbox.Scenario.Template.StageBased.html#c:instances/0"},{"type":"function","doc":"Returns list of of asset units.","title":"Runbox.Scenario.Template.StageBased.instances/1","ref":"Runbox.Scenario.Template.StageBased.html#instances/1"},{"type":"callback","doc":"","title":"Runbox.Scenario.Template.StageBased.subscriptions/0","ref":"Runbox.Scenario.Template.StageBased.html#c:subscriptions/0"},{"type":"function","doc":"Returns subscriptions defined in `template_module`.","title":"Runbox.Scenario.Template.StageBased.subscriptions/1","ref":"Runbox.Scenario.Template.StageBased.html#subscriptions/1"},{"type":"type","doc":"","title":"Runbox.Scenario.Template.StageBased.handle_asset_discovery_resp/0","ref":"Runbox.Scenario.Template.StageBased.html#t:handle_asset_discovery_resp/0"},{"type":"type","doc":"","title":"Runbox.Scenario.Template.StageBased.handle_message_resp/0","ref":"Runbox.Scenario.Template.StageBased.html#t:handle_message_resp/0"},{"type":"type","doc":"","title":"Runbox.Scenario.Template.StageBased.instances_def/0","ref":"Runbox.Scenario.Template.StageBased.html#t:instances_def/0"},{"type":"type","doc":"","title":"Runbox.Scenario.Template.StageBased.routing_key_definition/0","ref":"Runbox.Scenario.Template.StageBased.html#t:routing_key_definition/0"},{"type":"type","doc":"","title":"Runbox.Scenario.Template.StageBased.subscription_def/0","ref":"Runbox.Scenario.Template.StageBased.html#t:subscription_def/0"},{"type":"type","doc":"","title":"Runbox.Scenario.Template.StageBased.template_output/0","ref":"Runbox.Scenario.Template.StageBased.html#t:template_output/0"},{"type":"module","doc":"Defines supported scenario types.\n\nYou can use this in Manifest instead of hard-coding the particular value.\n\n```\n%Runbox.Scenario.Manifest{\n  type: Runbox.Scenario.Type.stage(),\n  ...\n}\n```","title":"Runbox.Scenario.Type","ref":"Runbox.Scenario.Type.html"},{"type":"function","doc":"Returns all available types.","title":"Runbox.Scenario.Type.all_types/0","ref":"Runbox.Scenario.Type.html#all_types/0"},{"type":"function","doc":"Simple scenario.\n\nSee `Runbox.Scenario.Simple` for details about this scenario type.","title":"Runbox.Scenario.Type.simple/0","ref":"Runbox.Scenario.Type.html#simple/0"},{"type":"function","doc":"Type for GenStage-based scenario runtime.","title":"Runbox.Scenario.Type.stage/0","ref":"Runbox.Scenario.Type.html#stage/0"},{"type":"type","doc":"","title":"Runbox.Scenario.Type.t/0","ref":"Runbox.Scenario.Type.html#t:t/0"},{"type":"module","doc":"Definition of action to be fired from UI.\n\nIn scenario list of `Runbox.Scenario.UIAction`s is defined and stored in asset\nmap entity. When action is fired from UI, scenario receives `Runbox.Scenario.UserAction`.","title":"Runbox.Scenario.UIAction","ref":"Runbox.Scenario.UIAction.html"},{"type":"function","doc":"Converts list of UI actions to actions property.\n\nMap in resulting list contains two properties:\n  * `action` - action (`topic`, `type`, `details`) encoded to JWT.\n  * `meta` - metadata (`name`, `description`) to be used in UI.\n\nResulting list of maps should be stored to some entity in asset map.","title":"Runbox.Scenario.UIAction.actions/1","ref":"Runbox.Scenario.UIAction.html#actions/1"},{"type":"type","doc":"UI action definition.\n\nProperties `topic`, `type` and `details` are serialized to JWT token and used for\nrouting of fired action. Property `topic` is name of raw topic where fired actions\nare pushed, property `type` is type to be used in `Runbox.Message` as `type` and\nproperty `details` is stored in resulting message `body` as `details` property.\nProperties `name` and `description` are only metadata for UI.","title":"Runbox.Scenario.UIAction.t/0","ref":"Runbox.Scenario.UIAction.html#t:t/0"},{"type":"module","doc":"Provides topology sorting capabilities.","title":"Runbox.Utils.TopologySort","ref":"Runbox.Utils.TopologySort.html"},{"type":"function","doc":"Sort the given network by its topology.\n\nThe network is given in the form of a list with elements `{node, subs}` where subs is a list of\nnodes the node is connected to. The function returns list of nodes sorted by the topology.\n\nThe sort is stable - original ordering of the elements is maintained where possible.","title":"Runbox.Utils.TopologySort.sort/2","ref":"Runbox.Utils.TopologySort.html#sort/2"},{"type":"task","doc":"Assembles an Altworx scenario release for the current project.\n\nThis task is to be called internally by `Mix.Tasks.Altworx.ScenarioRelease` from inside Docker.\n\n    $ MIX_ENV=prod mix altworx.scenario_release_internal","title":"mix altworx.scenario_release_internal","ref":"Mix.Tasks.Altworx.ScenarioReleaseInternal.html"},{"type":"task","doc":"* `--force` - forces recompilation\n  * `--no-archives-check` - does not check archive\n  * `--no-deps-check` - does not check dependencies\n  * `--no-elixir-version-check` - does not check Elixir version\n  * `--no-compile` - does not compile before assembling the release\n  * `--overwrite` - if there is an existing release version, overwrite it\n  * `--path` - the path of the release\n  * `--quiet` - does not write progress to the standard output\n  * `--version` - the version of the release","title":"Command line options - mix altworx.scenario_release_internal","ref":"Mix.Tasks.Altworx.ScenarioReleaseInternal.html#module-command-line-options"},{"type":"module","doc":"Runbox allows working with scenario releases. It can load scenario releases\nand provide information about the individual scenarios within them.","title":"Runbox","ref":"Runbox.html"},{"type":"function","doc":"Generates initial state for run component on run slave node.\n\nBecause function should be called from code implementing scenario type (for example stage based),\nreturned value is also specific for scenario type.","title":"Runbox.generate_state_for_run_component/2","ref":"Runbox.html#generate_state_for_run_component/2"},{"type":"function","doc":"Retrieves information about scenarios in a release.","title":"Runbox.get_release_scenarios/1","ref":"Runbox.html#get_release_scenarios/1"},{"type":"function","doc":"","title":"Runbox.get_runbox_version/0","ref":"Runbox.html#get_runbox_version/0"},{"type":"function","doc":"Lists all channels with templates defined for scenario, notification type and template type.","title":"Runbox.list_scenario_notification_channels/4","ref":"Runbox.html#list_scenario_notification_channels/4"},{"type":"function","doc":"Lists all template languages defined for scenario, notification type, template type and channel.","title":"Runbox.list_scenario_notification_languages/5","ref":"Runbox.html#list_scenario_notification_languages/5"},{"type":"function","doc":"Lists all template types of the specified notification.","title":"Runbox.list_scenario_notification_template_types/3","ref":"Runbox.html#list_scenario_notification_template_types/3"},{"type":"function","doc":"Lists all notification types for a scenario.","title":"Runbox.list_scenario_notification_types/2","ref":"Runbox.html#list_scenario_notification_types/2"},{"type":"function","doc":"Loads and evaluates templates specified by scenario, notification type, template type, channel\nand language.\n\nIf language is not defined or does not exist, `fallback_language` is used.","title":"Runbox.load_and_eval_scenario_notification_templates/8","ref":"Runbox.html#load_and_eval_scenario_notification_templates/8"},{"type":"function","doc":"Runs `on_start` function of scenario on run slave node.","title":"Runbox.run_on_start/2","ref":"Runbox.html#run_on_start/2"},{"type":"function","doc":"Saves entity state to state store running on run master node.\n\nThis call is called from slave to master.","title":"Runbox.save_entity/3","ref":"Runbox.html#save_entity/3"},{"type":"function","doc":"Starts new environment for run.\n\nEvery run runs on own slave node. Caller process is linked with supervisor started during\nthis call. Started supervisor will supervise all run components running on slave node.","title":"Runbox.start_link_run_environment/2","ref":"Runbox.html#start_link_run_environment/2"},{"type":"function","doc":"Starts slave node for release in `release_dir`.\n\nIf `app_env` is set, configured applications env variables are set on slave node\n(for example `ui_url` variable is needed to be set in `runbox` application env if\nnode is used for notifications evaluation, i.e. `app_env` must be set to\n`[runbox: [ui_url: \"https://some_url\"]]`).","title":"Runbox.start_release_node/2","ref":"Runbox.html#start_release_node/2"},{"type":"function","doc":"Starts scenario component on slave node.","title":"Runbox.start_run_component/3","ref":"Runbox.html#start_run_component/3"},{"type":"module","doc":"Decide if a message is a duplicate by comparing it to messages already seen.\n\nOnly messages with the biggest timestamp are remembered as \"seen\". If the\nactual message has lower timestamp than the messages in \"seen\", it is\nconsidered \"old\". If it has the exact same timestamp, it is compared for\nequality with all the \"seen\" messages and if it matches, it is considered\n\"duplicate\". In any other case, the message is considered \"new\".","title":"Runbox.Deduplicator","ref":"Runbox.Deduplicator.html"},{"type":"function","doc":"Decides if a given message is a duplicate or not.\n\nReturns a tuple with `{msg_condition, deduplicator_state}` where `msg_condition` can be one of:\n\n  * `:new` - message is new and was not yet seen.\n\n  * `:old` - message is older than the messages already seen.\n\n  * `:duplicate` - message is equal to one of the latest messages already seen.","title":"Runbox.Deduplicator.deduplicate/2","ref":"Runbox.Deduplicator.html#deduplicate/2"},{"type":"function","doc":"Initializes a deduplicator.\n\n`stream` is an enumerable (possibly empty), which contains messages with\ndescending timestamps - from newest to oldest. It is used to initialize\n\"seen\" messages from messages which were already processed.\n\n`extract_timestamp` is a function which returns a timestamp for a given\nmessage.","title":"Runbox.Deduplicator.new/2","ref":"Runbox.Deduplicator.html#new/2"},{"type":"opaque","doc":"","title":"Runbox.Deduplicator.t/0","ref":"Runbox.Deduplicator.html#t:t/0"},{"type":"module","doc":"Utilities for communicating with master node from slave node","title":"Runbox.Master","ref":"Runbox.Master.html"},{"type":"function","doc":"Calls a function on the master.\n\nCalls MFA on the `master`. See `:erpc.call/4` for details.","title":"Runbox.Master.call/4","ref":"Runbox.Master.html#call/4"},{"type":"module","doc":"Structure holding Runbox context of running slave node.","title":"Runbox.RunContext","ref":"Runbox.RunContext.html"},{"type":"type","doc":"Context of running run slave node.\n\n  * `:master_node` - master node of run\n  * `:node` - slave node of run\n  * `:sup_pid` - pid of supervisor (running on slave) for run components running on slave\n  * `:save_entity_mf` - callback definition (module and function name) for saving entity\n    state. Function is always called with 2 params: entity (`t:Runbox.StateStore.Entity.t/0`)\n    and savepoint timestamp.","title":"Runbox.RunContext.t/0","ref":"Runbox.RunContext.html#t:t/0"},{"type":"module","doc":"Module contains common functions called on run slave node.","title":"Runbox.RunNode","ref":"Runbox.RunNode.html"},{"type":"function","doc":"Starts run component under run scenario.","title":"Runbox.RunNode.start_run_component/3","ref":"Runbox.RunNode.html#start_run_component/3"},{"type":"function","doc":"Starts supervisor for run.\n\nAll components of run running on slave node will run under this supervisor.\nOneForAll strategy with 0 max restart stops supervisor if any component crashes.\nBecause run manager process running on master is linked with this supervisor,\nit also crashes (and it what we want).","title":"Runbox.RunNode.start_run_sup/1","ref":"Runbox.RunNode.html#start_run_sup/1"},{"type":"module","doc":"Structure holding context of starting run.\n\nThis structure is passed as param to every run component start.","title":"Runbox.RunStartContext","ref":"Runbox.RunStartContext.html"},{"type":"function","doc":"Adds pid of component identified by `name` to context.","title":"Runbox.RunStartContext.add_component/3","ref":"Runbox.RunStartContext.html#add_component/3"},{"type":"function","doc":"Returns pid of component identitfied by `name`.","title":"Runbox.RunStartContext.component_pid/2","ref":"Runbox.RunStartContext.html#component_pid/2"},{"type":"type","doc":"Currently context holds only mapping of names to pids of already started components.","title":"Runbox.RunStartContext.t/0","ref":"Runbox.RunStartContext.html#t:t/0"},{"type":"behaviour","doc":"Component network builds and validates run component dependency network.\n\nTo build this network, module uses dependencies defined by the scenario.","title":"Runbox.Runtime.ComponentNetwork","ref":"Runbox.Runtime.ComponentNetwork.html"},{"type":"callback","doc":"","title":"Runbox.Runtime.ComponentNetwork.convert_to_network/1","ref":"Runbox.Runtime.ComponentNetwork.html#c:convert_to_network/1"},{"type":"function","doc":"Creates base network with only template nodes","title":"Runbox.Runtime.ComponentNetwork.convert_to_network/2","ref":"Runbox.Runtime.ComponentNetwork.html#convert_to_network/2"},{"type":"callback","doc":"","title":"Runbox.Runtime.ComponentNetwork.create/2","ref":"Runbox.Runtime.ComponentNetwork.html#c:create/2"},{"type":"function","doc":"Main function. Given list of template modules,\nexpands them into components network with input streams, timezips etc...","title":"Runbox.Runtime.ComponentNetwork.create/3","ref":"Runbox.Runtime.ComponentNetwork.html#create/3"},{"type":"function","doc":"* `:direct_ticking` - if true (default), creates network using direct ticking","title":"Options - Runbox.Runtime.ComponentNetwork.create/3","ref":"Runbox.Runtime.ComponentNetwork.html#create/3-options"},{"type":"callback","doc":"","title":"Runbox.Runtime.ComponentNetwork.input_topics/1","ref":"Runbox.Runtime.ComponentNetwork.html#c:input_topics/1"},{"type":"function","doc":"Return all input topic names that are being subscribed to in component network.","title":"Runbox.Runtime.ComponentNetwork.input_topics/2","ref":"Runbox.Runtime.ComponentNetwork.html#input_topics/2"},{"type":"function","doc":"","title":"Runbox.Runtime.ComponentNetwork.topology_sort/2","ref":"Runbox.Runtime.ComponentNetwork.html#topology_sort/2"},{"type":"type","doc":"","title":"Runbox.Runtime.ComponentNetwork.component/0","ref":"Runbox.Runtime.ComponentNetwork.html#t:component/0"},{"type":"type","doc":"","title":"Runbox.Runtime.ComponentNetwork.t/0","ref":"Runbox.Runtime.ComponentNetwork.html#t:t/0"},{"type":"module","doc":"Output Action for internal use by Runtime","title":"Runbox.Runtime.OutputAction","ref":"Runbox.Runtime.OutputAction.html"},{"type":"function","doc":"Creates a new output action.\n\nTo produce a valid output action, the `params` have to be one of the accepted\nstructs. See the documentation of the particular structs for details.\n\nIf the `params` are not one of the accepted structs or some of the fields do\nnot pass validation, it returns an output action with body set to\n`t:Runbox.Scenario.OutputAction.BadOutputAction.t/0` with original params\ninside.","title":"Runbox.Runtime.OutputAction.new/4","ref":"Runbox.Runtime.OutputAction.html#new/4"},{"type":"function","doc":"Returns `true` if valid output action body.","title":"Runbox.Runtime.OutputAction.oa_body?/1","ref":"Runbox.Runtime.OutputAction.html#oa_body?/1"},{"type":"function","doc":"Returns `true` if valid output action.","title":"Runbox.Runtime.OutputAction.valid?/1","ref":"Runbox.Runtime.OutputAction.html#valid?/1"},{"type":"type","doc":"Runtime Output Action","title":"Runbox.Runtime.OutputAction.t/0","ref":"Runbox.Runtime.OutputAction.html#t:t/0"},{"type":"module","doc":"Component network builds and validates template dependency network for Simple scenario runtime.\n\nThe component network for Simple scenarios is quite simple.\n* first you have input topics, usually followed by tick timezip for each of them\n* following is a timezip zipping all input topics into a single stream\n* the stream is processed by the template which is the heart of the scenario, it produces a stream of output actions\n* finally there's an output sink executing the output actions","title":"Runbox.Runtime.Simple.ComponentNetwork","ref":"Runbox.Runtime.Simple.ComponentNetwork.html"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.ComponentNetwork.component/0","ref":"Runbox.Runtime.Simple.ComponentNetwork.html#t:component/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.ComponentNetwork.input/0","ref":"Runbox.Runtime.Simple.ComponentNetwork.html#t:input/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.ComponentNetwork.output_sink/0","ref":"Runbox.Runtime.Simple.ComponentNetwork.html#t:output_sink/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.ComponentNetwork.t/0","ref":"Runbox.Runtime.Simple.ComponentNetwork.html#t:t/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.ComponentNetwork.template/0","ref":"Runbox.Runtime.Simple.ComponentNetwork.html#t:template/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.ComponentNetwork.tick_timezip/0","ref":"Runbox.Runtime.Simple.ComponentNetwork.html#t:tick_timezip/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.ComponentNetwork.timezip/0","ref":"Runbox.Runtime.Simple.ComponentNetwork.html#t:timezip/0"},{"type":"module","doc":"Template carrier for Simple scenario.\n\nGenStage handling Simple scenario template. It carries the state of the template and handles\n`init` and `handle_message` callbacks. It expects to consume messages from input streams and\nproduces output-stream-ready events.","title":"Runbox.Runtime.Simple.TemplateCarrier","ref":"Runbox.Runtime.Simple.TemplateCarrier.html"},{"type":"function","doc":"Returns component name.","title":"Runbox.Runtime.Simple.TemplateCarrier.component_name/0","ref":"Runbox.Runtime.Simple.TemplateCarrier.html#component_name/0"},{"type":"function","doc":"Returns state entity of a running template carrier.","title":"Runbox.Runtime.Simple.TemplateCarrier.get_state_entity/1","ref":"Runbox.Runtime.Simple.TemplateCarrier.html#get_state_entity/1"},{"type":"function","doc":"Starts the GenStage.","title":"Runbox.Runtime.Simple.TemplateCarrier.start_link/3","ref":"Runbox.Runtime.Simple.TemplateCarrier.html#start_link/3"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.TemplateCarrier.origin/0","ref":"Runbox.Runtime.Simple.TemplateCarrier.html#t:origin/0"},{"type":"type","doc":"Stage output (input for the next stage) is either a tick message or an output action","title":"Runbox.Runtime.Simple.TemplateCarrier.stage_outputs/0","ref":"Runbox.Runtime.Simple.TemplateCarrier.html#t:stage_outputs/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.TemplateCarrier.state_entity/0","ref":"Runbox.Runtime.Simple.TemplateCarrier.html#t:state_entity/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.TemplateCarrier.timestamp/0","ref":"Runbox.Runtime.Simple.TemplateCarrier.html#t:timestamp/0"},{"type":"module","doc":"State that is periodically persisted to savepoints and loaded when run continues.","title":"Runbox.Runtime.Simple.TemplateCarrier.Snapshot","ref":"Runbox.Runtime.Simple.TemplateCarrier.Snapshot.html"},{"type":"function","doc":"Upgrades a possibly old version of the struct to the current version.\n\nThe upgrade process is based on steps. Each step applies transformation of the state\nrelated to the changes in a specific version. It is similar to database migrations.\nThis function executes all the upgrade steps since the version of the given struct\nuntil the struct is upgraded to the current version.","title":"Runbox.Runtime.Simple.TemplateCarrier.Snapshot.upgrade/1","ref":"Runbox.Runtime.Simple.TemplateCarrier.Snapshot.html#upgrade/1"},{"type":"function","doc":"For example, if version 42.0.0 adds the field `:new_field`, then you should create\na corresponding clause in the `do_upgrade_step/1` function. This clause should upgrade\nthe state from the last preceding version (e.g., 41.1.0) and update the `:runbox_version`\nfield. Like so:\n\n    defp do_upgrade_step(%{runbox_version: %{major: major}} = snapshot) when major < 42 do\n      snapshot\n      |> Map.put(:new_field, :some_value)\n      |> Map.put(:runbox_version, Version.parse!(\"42.0.0\"))\n    end\n\nPlace this new clause right before the catch-all clause (`defp do_upgrade_step(snapshot)`).","title":"Adding a new upgrade step - Runbox.Runtime.Simple.TemplateCarrier.Snapshot.upgrade/1","ref":"Runbox.Runtime.Simple.TemplateCarrier.Snapshot.html#upgrade/1-adding-a-new-upgrade-step"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.TemplateCarrier.Snapshot.t/0","ref":"Runbox.Runtime.Simple.TemplateCarrier.Snapshot.html#t:t/0"},{"type":"module","doc":"State of a running template carrier for a Simple scenario.","title":"Runbox.Runtime.Simple.TemplateCarrier.State","ref":"Runbox.Runtime.Simple.TemplateCarrier.State.html"},{"type":"type","doc":"","title":"Runbox.Runtime.Simple.TemplateCarrier.State.t/0","ref":"Runbox.Runtime.Simple.TemplateCarrier.State.html#t:t/0"},{"type":"module","doc":"Timezip component responsible for merging multiple message streams for Simple runtime.\n\nIn a Simple scenario, there is only need for a Timezip if reading from multiple input topics. Then\nwe need to merge these streams into a single stream for the template.","title":"Runbox.Runtime.Simple.Timezip","ref":"Runbox.Runtime.Simple.Timezip.html"},{"type":"function","doc":"Returns component name.\n\nThere are no parameters, since in Simple scenario runtime there is at most one Timezip.","title":"Runbox.Runtime.Simple.Timezip.component_name/0","ref":"Runbox.Runtime.Simple.Timezip.html#component_name/0"},{"type":"function","doc":"Starts the GenStage.","title":"Runbox.Runtime.Simple.Timezip.start_link/3","ref":"Runbox.Runtime.Simple.Timezip.html#start_link/3"},{"type":"module","doc":"Component network builds and validates template dependency network for StageBased scenario runtime.\n\nTo build this network, module uses `c:Runbox.Scenario.Template.StageBased.subscriptions/0`.\nOnly templates are defined at first, logic than derive input-topic nodes, timezip where one\ntemplate subscribes to two stream, and output sink in the end.","title":"Runbox.Runtime.Stage.ComponentNetwork","ref":"Runbox.Runtime.Stage.ComponentNetwork.html"},{"type":"function","doc":"Creates base network with only template nodes","title":"Runbox.Runtime.Stage.ComponentNetwork.convert_to_network/1","ref":"Runbox.Runtime.Stage.ComponentNetwork.html#convert_to_network/1"},{"type":"function","doc":"Main function. Given list of template modules,\nexpands them into components network with input streams, timezips etc...","title":"Runbox.Runtime.Stage.ComponentNetwork.create/2","ref":"Runbox.Runtime.Stage.ComponentNetwork.html#create/2"},{"type":"function","doc":"* `:direct_ticking` - if true (default), creates network using direct ticking","title":"Options - Runbox.Runtime.Stage.ComponentNetwork.create/2","ref":"Runbox.Runtime.Stage.ComponentNetwork.html#create/2-options"},{"type":"function","doc":"Return all input topic names that are being subscribed to in component network.","title":"Runbox.Runtime.Stage.ComponentNetwork.input_topics/1","ref":"Runbox.Runtime.Stage.ComponentNetwork.html#input_topics/1"},{"type":"function","doc":"Sorts components (keys) of component network topologicaly, but according to backward edges,\nbecause edge means subscription relation (not comes after but comes before).\nSo instead of indegree implementation use outdegee. Its wrath of kahn's algorithm.","title":"Runbox.Runtime.Stage.ComponentNetwork.topology_sort/1","ref":"Runbox.Runtime.Stage.ComponentNetwork.html#topology_sort/1"},{"type":"type","doc":"Full representation of particular component.","title":"Runbox.Runtime.Stage.ComponentNetwork.component/0","ref":"Runbox.Runtime.Stage.ComponentNetwork.html#t:component/0"},{"type":"type","doc":"Part of component identifier.","title":"Runbox.Runtime.Stage.ComponentNetwork.component_type/0","ref":"Runbox.Runtime.Stage.ComponentNetwork.html#t:component_type/0"},{"type":"type","doc":"Subscription edge defined with destination and props.","title":"Runbox.Runtime.Stage.ComponentNetwork.subscription/0","ref":"Runbox.Runtime.Stage.ComponentNetwork.html#t:subscription/0"},{"type":"type","doc":"Additional properties of subscription/edge. E.g. selector for Broadcast dispatcher.","title":"Runbox.Runtime.Stage.ComponentNetwork.subscription_properties/0","ref":"Runbox.Runtime.Stage.ComponentNetwork.html#t:subscription_properties/0"},{"type":"type","doc":"Component network is adjacency graph with subscription edges.","title":"Runbox.Runtime.Stage.ComponentNetwork.t/0","ref":"Runbox.Runtime.Stage.ComponentNetwork.html#t:t/0"},{"type":"module","doc":"StageBased Template emulator.\n\nThis is similar to `Runbox.Runtime.Stage.TemplateImpl` but designed specifically for the\nin-process emulation via `Runbox.Runtime.Stage.Emulator`.\n\nIf emulating StageBased scenario, do not use this module directly, rather use\n`Runbox.Runtime.Stage.Emulator` to emulate the whole scenario.","title":"Runbox.Runtime.Stage.Emulator.Template","ref":"Runbox.Runtime.Stage.Emulator.Template.html"},{"type":"function","doc":"Returns whether this template requires deduplication on the inputs.\n\nIn a StageBased scenario a deduplication is performed by a Timezip component. So we need to\nperform deduplication only in places where there would be a Timezip. Other places should happily\nreceive duplicated messages.","title":"Runbox.Runtime.Stage.Emulator.Template.deduplicate_inputs?/1","ref":"Runbox.Runtime.Stage.Emulator.Template.html#deduplicate_inputs?/1"},{"type":"function","doc":"Create a struct representing an emulated template.\n\nThe struct holds all necessary information about the template, including its current state. This\nstruct is used during the emulation, so you generally start by creating this struct via this\nfunction.","title":"Runbox.Runtime.Stage.Emulator.Template.from_module/1","ref":"Runbox.Runtime.Stage.Emulator.Template.html#from_module/1"},{"type":"function","doc":"Handle a message in the template.\n\nThis involves the following:\n- use the routing part of the subscriptions to find all units that should handle the message\n- handle the message in all these units via the\n  `c:Runbox.Scenario.Template.StageBased.handle_message/2` callback\n- if no unit was found trigger `c:Runbox.Scenario.Template.StageBased.handle_asset_discovery/1` to\n  potentially create a new unit\n\nNote that timeouts are routed directly to the unit that created them, they are not subjected to\nthe same routing as other messages which come from input topics or other templates.","title":"Runbox.Runtime.Stage.Emulator.Template.handle_message/2","ref":"Runbox.Runtime.Stage.Emulator.Template.html#handle_message/2"},{"type":"function","doc":"Initialize static defined units.\n\nStatic units are defined in `c:Runbox.Scenario.Template.StageBased.instances/0` callback. The\nunits are created and initialized via the `c:Runbox.Scenario.Template.StageBased.init/2` callback.","title":"Runbox.Runtime.Stage.Emulator.Template.initialize_static_units/2","ref":"Runbox.Runtime.Stage.Emulator.Template.html#initialize_static_units/2"},{"type":"function","doc":"Returns input topics the template module is subscribed to.","title":"Runbox.Runtime.Stage.Emulator.Template.input_topics/1","ref":"Runbox.Runtime.Stage.Emulator.Template.html#input_topics/1"},{"type":"function","doc":"Re-initialize a template.\n\nThis step is necessary when a template is reconstructed from a savepoint or after a similar\ntransformation which might corrupt internal template state.\n\nThis is required, because this module depends on `Runbox.Runtime.Stage.UnitRegistry` which keeps\nanonymous functions in its state. Such function references might not survive a persistence (e.g.\ninto ETF) and need to be reconstructed.","title":"Runbox.Runtime.Stage.Emulator.Template.reinitialize_template/1","ref":"Runbox.Runtime.Stage.Emulator.Template.html#reinitialize_template/1"},{"type":"function","doc":"Determines whether a template should process a message.\n\nTemplate should process a message either when the message matches one of its subscriptions or when\nthe message is a planned timeout for a unit of this template. A part of this information is where\nis the message coming from - a specific logical topic or template.","title":"Runbox.Runtime.Stage.Emulator.Template.should_handle_message?/3","ref":"Runbox.Runtime.Stage.Emulator.Template.html#should_handle_message?/3"},{"type":"function","doc":"Sorts a list of templates in a topological order by their subscriptions.\n\nTemplates can depend on each other but these dependencies can never form a loop. Therefore, we can\nsort templates using these dependencies into a topological order. This gives us the templates in\nthe order in which they can process messages. Then a latter template might consume outputs of a\nformer template, but never the other way round.\n\nThis also ensures the order is the same as in StageBased sorted component network. That is used to\nmake Timezip emit outputs in a deterministic manner. Since there is no Timezip in the emulation we\nrely solely on the order of the templates to achieve the same order of outputs.","title":"Runbox.Runtime.Stage.Emulator.Template.topological_sort/1","ref":"Runbox.Runtime.Stage.Emulator.Template.html#topological_sort/1"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.Emulator.Template.scenario_output/0","ref":"Runbox.Runtime.Stage.Emulator.Template.html#t:scenario_output/0"},{"type":"opaque","doc":"","title":"Runbox.Runtime.Stage.Emulator.Template.t/0","ref":"Runbox.Runtime.Stage.Emulator.Template.html#t:t/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.Emulator.Template.template_output/0","ref":"Runbox.Runtime.Stage.Emulator.Template.html#t:template_output/0"},{"type":"module","doc":"Transform valid component network to gen_stage definitions.","title":"Runbox.Runtime.Stage.Sandbox.StartRecipe","ref":"Runbox.Runtime.Stage.Sandbox.StartRecipe.html"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.Sandbox.StartRecipe.component_name/1","ref":"Runbox.Runtime.Stage.Sandbox.StartRecipe.html#component_name/1"},{"type":"function","doc":"Transform valid component network to gen_stage definitions.","title":"Runbox.Runtime.Stage.Sandbox.StartRecipe.create/1","ref":"Runbox.Runtime.Stage.Sandbox.StartRecipe.html#create/1"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.Sandbox.StartRecipe.component/0","ref":"Runbox.Runtime.Stage.Sandbox.StartRecipe.html#t:component/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.Sandbox.StartRecipe.component_network/0","ref":"Runbox.Runtime.Stage.Sandbox.StartRecipe.html#t:component_network/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.Sandbox.StartRecipe.subscription/0","ref":"Runbox.Runtime.Stage.Sandbox.StartRecipe.html#t:subscription/0"},{"type":"module","doc":"Selector function generator which is used to filter messages between GenStage phases.\n\nStage based runtime uses BroadcastDispatcher to dispatch messages between TemplateCarriers,\nthis kind of dispatcher support message selector function to be passed as part of GenStage\nsubscription. This module is used to generate this function based on template configuration.","title":"Runbox.Runtime.Stage.SelectorBuilder","ref":"Runbox.Runtime.Stage.SelectorBuilder.html"},{"type":"function","doc":"Builds GenStages BroadcastDispatcher filter function based on template configuration.\n\nFilter function is produced based on `c:Runbox.Scenario.Template.StageBased.subscriptions/0`\ncallback result.\n\nSubscription is defined as {message_type, _routing_rule}, we use only message_type part\nof subscription. All message types are stored in produced function clojure and dispatched\nmessage type is matched against them.\n\nException to the message type matching are output actions and tick messages which are\nalways matched.","title":"Runbox.Runtime.Stage.SelectorBuilder.build/1","ref":"Runbox.Runtime.Stage.SelectorBuilder.html#build/1"},{"type":"module","doc":"Module generates state of run components running on slave node.","title":"Runbox.Runtime.Stage.StateGenerator","ref":"Runbox.Runtime.Stage.StateGenerator.html"},{"type":"function","doc":"Generates state for component template.","title":"Runbox.Runtime.Stage.StateGenerator.generate_for_component/2","ref":"Runbox.Runtime.Stage.StateGenerator.html#generate_for_component/2"},{"type":"module","doc":"Module implements producer_consumer GenStage behaviour, scenario template is used to handle events\n\nEvery scenario template is represented in stage based runtime by this module. TemplateCarrier uses\n`Runbox.Runtime.Stage.UnitRegistry` as its state (packed in `Runbox.StateStore.Entity`\nenvelope for `StateStore` usage). `UnitRegistry` provides functionality to lookup units via\nsubscriptions defined in scenario template.\n\nIf unit is located for given message,\n`c:Runbox.Scenario.Template.StageBased.handle_message/2` is used to handle this message. When\nthere is no registered unit for given routing key,\n`c:Runbox.Scenario.Template.StageBased.handle_asset_discovery/1` is used instead.\n\nThe state is updated in `Runbox.StateStore.Entity` with\n`Runbox.StateStore.Entity.update_state/3` after every message is processed. Additionally,\nbefore a message is processed it is confirmed that all messages with lower timestamps were\nprocessed. This is done with `Runbox.StateStore.Entity.ack_processed_time/3` and can result in\nsaving the state in state store.","title":"Runbox.Runtime.Stage.TemplateCarrier","ref":"Runbox.Runtime.Stage.TemplateCarrier.html"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.TemplateCarrier.component_name/1","ref":"Runbox.Runtime.Stage.TemplateCarrier.html#component_name/1"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.TemplateCarrier.start_link/3","ref":"Runbox.Runtime.Stage.TemplateCarrier.html#start_link/3"},{"type":"module","doc":"Module encapsulates scenario template callback calls.\n\nScenario defines `Runbox.Runtime.Stage.Unit` as state carrier, these units are organized\ninto `Runbox.Runtime.Stage.UnitRegistry` which is helper structure to route messages between\nunits managed by one template.\n\nMessages are handled via `Runbox.Scenario.Template.StageBased` callbacks. We distinguish two\ntypes of messages = message to existing unit and message to new unit (asset discovery).\n\nMessages that doesn't have any unit registered and thus found to handle them, we call this\nsituation asset discovery and use `c:Runbox.Scenario.Template.StageBased.handle_asset_discovery/1`\ncallback to handle these message. New unit may or may not be created here, but when it does,\nit needs to be initialized via `c:Runbox.Scenario.Template.StageBased.init/2` callback.\n\nWhen there is at least one unit which is registered to the message, we use callback\n`c:Runbox.Scenario.Template.StageBased.handle_message/2`. Unit may be updated or stopped here.\n\nWe support timeouts to be defined. These timeouts are triggered via `external` time, which is\nderived from incoming message timestamps. Timeouts are registered for one unit.\n\nSome output actions produced by any of template callback should be further updated. Events\nrequire filling `run id` and notifications require `scenario id` to be filled.","title":"Runbox.Runtime.Stage.TemplateImpl","ref":"Runbox.Runtime.Stage.TemplateImpl.html"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.TemplateImpl.handle_message/6","ref":"Runbox.Runtime.Stage.TemplateImpl.html#handle_message/6"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.TemplateImpl.init_state/5","ref":"Runbox.Runtime.Stage.TemplateImpl.html#init_state/5"},{"type":"module","doc":"Timezip component responsible for merging multiple message streams using time for Stage runtime.","title":"Runbox.Runtime.Stage.Timezip","ref":"Runbox.Runtime.Stage.Timezip.html"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.Timezip.component_name/1","ref":"Runbox.Runtime.Stage.Timezip.html#component_name/1"},{"type":"function","doc":"","title":"Runbox.Runtime.Stage.Timezip.start_link/3","ref":"Runbox.Runtime.Stage.Timezip.html#start_link/3"},{"type":"module","doc":"Unit register used in stage based runtime to manage state of units.\n\nUnitRegistry is configured via `c:Runbox.Scenario.Template.StageBased.subscriptions/0` callback\nresult. Callback returns template subscriptions, each subscription is defined as\n{message_type, routing_rule}. Unit registry contains one register per configured message_type.\nRouting rule is used to store unit and later on lookup this unit when needed.\n\nUnits are being registered using unit attributes as defined in routing_rule. Unit lookup uses\nrouting_rule to parse message body and the result compares to stored unit attributes.","title":"Runbox.Runtime.Stage.UnitRegistry","ref":"Runbox.Runtime.Stage.UnitRegistry.html"},{"type":"function","doc":"Returns unit with given id.","title":"Runbox.Runtime.Stage.UnitRegistry.get_unit/2","ref":"Runbox.Runtime.Stage.UnitRegistry.html#get_unit/2"},{"type":"function","doc":"Initialize UnitRegistry non-persisted fields.\n\nThis needs to be called right before starting new unit registry (done in `new/1` automatically)\nand also after loading UnitRegistry from savepoint. This is needed because not all fields are\npersisted into a savepoint.","title":"Runbox.Runtime.Stage.UnitRegistry.init/1","ref":"Runbox.Runtime.Stage.UnitRegistry.html#init/1"},{"type":"function","doc":"Searches for registered unit in the unit registry.","title":"Runbox.Runtime.Stage.UnitRegistry.lookup/2","ref":"Runbox.Runtime.Stage.UnitRegistry.html#lookup/2"},{"type":"function","doc":"Creates new unit registry.","title":"Runbox.Runtime.Stage.UnitRegistry.new/1","ref":"Runbox.Runtime.Stage.UnitRegistry.html#new/1"},{"type":"function","doc":"Returns first reached timeout and removes it from unit registry in one operation.","title":"Runbox.Runtime.Stage.UnitRegistry.pop_reached_timeout/2","ref":"Runbox.Runtime.Stage.UnitRegistry.html#pop_reached_timeout/2"},{"type":"function","doc":"Registers given unit to the unit registry.","title":"Runbox.Runtime.Stage.UnitRegistry.register/2","ref":"Runbox.Runtime.Stage.UnitRegistry.html#register/2"},{"type":"function","doc":"Registers timeout.","title":"Runbox.Runtime.Stage.UnitRegistry.register_timeout/4","ref":"Runbox.Runtime.Stage.UnitRegistry.html#register_timeout/4"},{"type":"function","doc":"Reregisters unit in the unit registry.\n\nFunction updates alternative registers using new version of unit attributes. Timeouts will remain\nuntouched.","title":"Runbox.Runtime.Stage.UnitRegistry.reregister/2","ref":"Runbox.Runtime.Stage.UnitRegistry.html#reregister/2"},{"type":"function","doc":"Unregisters given unit from the unit registry.","title":"Runbox.Runtime.Stage.UnitRegistry.unregister/2","ref":"Runbox.Runtime.Stage.UnitRegistry.html#unregister/2"},{"type":"function","doc":"Updates unit in unit registry.","title":"Runbox.Runtime.Stage.UnitRegistry.update/2","ref":"Runbox.Runtime.Stage.UnitRegistry.html#update/2"},{"type":"function","doc":"Upgrades a possibly old version of the struct to the current version.\n\nThe upgrade process is based on steps. Each step applies transformation of the state\nrelated to the changes in a specific version. It is similar to database migrations.\nThis function executes all the upgrade steps since the version of the given struct\nuntil the struct is upgraded to the current version.","title":"Runbox.Runtime.Stage.UnitRegistry.upgrade/1","ref":"Runbox.Runtime.Stage.UnitRegistry.html#upgrade/1"},{"type":"function","doc":"For example, if version 42.0.0 adds the field `:new_field`, then you should create\na corresponding clause in the `do_upgrade_step/1` function. This clause should upgrade\nthe state from the last preceding version (e.g., 41.1.0) and update the `:runbox_version`\nfield. Like so:\n\n    defp do_upgrade_step(%{runbox_version: %{major: major}} = snapshot) when major < 42 do\n      snapshot\n      |> Map.put(:new_field, :some_value)\n      |> Map.put(:runbox_version, Version.parse!(\"42.0.0\"))\n    end\n\nPlace this new clause right before the catch-all clause (`defp do_upgrade_step(snapshot)`).","title":"Adding a new upgrade step - Runbox.Runtime.Stage.UnitRegistry.upgrade/1","ref":"Runbox.Runtime.Stage.UnitRegistry.html#upgrade/1-adding-a-new-upgrade-step"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.UnitRegistry.msg_parser/0","ref":"Runbox.Runtime.Stage.UnitRegistry.html#t:msg_parser/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.UnitRegistry.path/0","ref":"Runbox.Runtime.Stage.UnitRegistry.html#t:path/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.UnitRegistry.routing_key/0","ref":"Runbox.Runtime.Stage.UnitRegistry.html#t:routing_key/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.UnitRegistry.routing_key_def/0","ref":"Runbox.Runtime.Stage.UnitRegistry.html#t:routing_key_def/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.UnitRegistry.t/0","ref":"Runbox.Runtime.Stage.UnitRegistry.html#t:t/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.UnitRegistry.unit_id/0","ref":"Runbox.Runtime.Stage.UnitRegistry.html#t:unit_id/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.UnitRegistry.unit_register/0","ref":"Runbox.Runtime.Stage.UnitRegistry.html#t:unit_register/0"},{"type":"module","doc":"Module manages alternative registry state.\n\nEvery unit is registered only once per alternative registry. More than one unit can be registered\nwith the same key.","title":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry","ref":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.html"},{"type":"function","doc":"Get registered unit ids","title":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.lookup/2","ref":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.html#lookup/2"},{"type":"function","doc":"Returns new alternative registry","title":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.new/0","ref":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.html#new/0"},{"type":"function","doc":"Registers unit in alternative reg with given key","title":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.register_unit/3","ref":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.html#register_unit/3"},{"type":"function","doc":"Unregisters unit from alternative reg","title":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.unregister_unit/2","ref":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.html#unregister_unit/2"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.registration_key/0","ref":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.html#t:registration_key/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.t/0","ref":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.html#t:t/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.unit_id/0","ref":"Runbox.Runtime.Stage.UnitRegistry.AlternativeRegistry.html#t:unit_id/0"},{"type":"module","doc":"Builder of msg_parser and unit_register fns used in `Runbox.Runtime.Stage.UnitRegistry`\n\n- message_parser is lambda function used to parse `Runbox.Message` and creates key to find unit\nin unit registry\n- unit_register is lambda function used to parse `Runbox.Runtime.Stage.Unit` and creates key\nto store unit in unit registry","title":"Runbox.Runtime.Stage.UnitRegistry.RoutingKeyBuilder","ref":"Runbox.Runtime.Stage.UnitRegistry.RoutingKeyBuilder.html"},{"type":"function","doc":"Gets routing key definition as argument and returns msg_parser lambda function\n\nLambda takes `Runbox.Message` as argument and returns list of message body values based on\nprovided routing key definition.\n\niex> msg_parser_fn =\n...> RoutingKeyBuilder.build_msg_parser({:=, [[:foo]], [[:bar]]})\n...> msg_parser_fn.(%Message{body: %{foo: \"message_body_key_value\"}})\n[\"message_body_key_value\"]\n\niex> msg_parser_fn =\n...> RoutingKeyBuilder.build_msg_parser({:in, [[:foo]], [[:bar]]})\n...> msg_parser_fn.(%Message{body: %{foo: \"message_body_key_value\"}})\n[\"message_body_key_value\"]","title":"Runbox.Runtime.Stage.UnitRegistry.RoutingKeyBuilder.build_msg_parser/1","ref":"Runbox.Runtime.Stage.UnitRegistry.RoutingKeyBuilder.html#build_msg_parser/1"},{"type":"function","doc":"Gets routing key definition as argument and returns unit_register lambda function\n\nLambda takes alternative registers (unit indexes), alternative register name,\n`Runbox.Runtime.Stage.Unit` as arguments and returns updated alternative registers based on\nprovided routing key definition.\n\niex> unit_register_fn =\n...> RoutingKeyBuilder.build_unit_register({:=, [[:foo]], [[:bar]]})\n...> unit = %Unit{id: \"unit_id\", attributes: %{bar: \"attribute_key_value\"}}\n...> unit_register_fn.(%{\"alt_reg\" => AlternativeRegistry.new()}, \"alt_reg\", unit)\n%{\"alt_reg\" => %{[\"attribute_key_value\"] => [\"unit_id\"]}}\n\niex> unit_register_fn =\n...> RoutingKeyBuilder.build_unit_register({:in, [[:foo]], [[:bar]]})\n...> unit = %Unit{id: \"unit_id\", attributes: %{bar: [\"value_1\", \"value_2\"]}}\n...> unit_register_fn.(%{\"alt_reg\" => AlternativeRegistry.new()}, \"alt_reg\", unit)\n%{\"alt_reg\" => %{[\"value_1\"] => [\"unit_id\"], [\"value_2\"] => [\"unit_id\"]}}","title":"Runbox.Runtime.Stage.UnitRegistry.RoutingKeyBuilder.build_unit_register/1","ref":"Runbox.Runtime.Stage.UnitRegistry.RoutingKeyBuilder.html#build_unit_register/1"},{"type":"module","doc":"Buffering multi-queue data structure for Timezip.\n\nIf the structure would be heap, the complexity would be Olog(N) where N is number of messages\n(topped by max demand).\nWith multi-queue operation is ~ O(M) where M is number of producers.\n\nThe produces / queues are ordered. Messages in the same timestamp across different queues are\nproduced in this order.","title":"Runbox.Runtime.Timezip.MultiQueue","ref":"Runbox.Runtime.Timezip.MultiQueue.html"},{"type":"function","doc":"Adds new queue to MultiQueue.\n\nAdds a new empty queue at the end.","title":"Runbox.Runtime.Timezip.MultiQueue.add_queue/2","ref":"Runbox.Runtime.Timezip.MultiQueue.html#add_queue/2"},{"type":"function","doc":"Main logic, will return messages from all queues zipped/sorted by min timestamp.\n\nWill return as many messages as possible, until some queue is empty.\n\nIn returned tuple, returned values means:\n - *multi_queue:* multi-queue without emitted messages\n\n - *demand_distribution:* map with `id -> integer()`, meaning how many messages in output\n   originates from particular queue id, can be used for manual demand handling in genstage\n\n - *emitted:* returned messages from multiqueue front","title":"Runbox.Runtime.Timezip.MultiQueue.dequeue_all/2","ref":"Runbox.Runtime.Timezip.MultiQueue.html#dequeue_all/2"},{"type":"function","doc":"Adds messages for queue id to multiqueue.","title":"Runbox.Runtime.Timezip.MultiQueue.enqueue/3","ref":"Runbox.Runtime.Timezip.MultiQueue.html#enqueue/3"},{"type":"function","doc":"","title":"Runbox.Runtime.Timezip.MultiQueue.new/0","ref":"Runbox.Runtime.Timezip.MultiQueue.html#new/0"},{"type":"function","doc":"Removes queue and its content from multi_queue","title":"Runbox.Runtime.Timezip.MultiQueue.remove_queue/2","ref":"Runbox.Runtime.Timezip.MultiQueue.html#remove_queue/2"},{"type":"type","doc":"","title":"Runbox.Runtime.Timezip.MultiQueue.demand_distribution/0","ref":"Runbox.Runtime.Timezip.MultiQueue.html#t:demand_distribution/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Timezip.MultiQueue.queue_id/0","ref":"Runbox.Runtime.Timezip.MultiQueue.html#t:queue_id/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Timezip.MultiQueue.t/0","ref":"Runbox.Runtime.Timezip.MultiQueue.html#t:t/0"},{"type":"module","doc":"Holds information about scenario.","title":"Runbox.Scenario","ref":"Runbox.Scenario.html"},{"type":"function","doc":"","title":"Runbox.Scenario.get_impl_for/1","ref":"Runbox.Scenario.html#get_impl_for/1"},{"type":"function","doc":"","title":"Runbox.Scenario.manifest_module?/1","ref":"Runbox.Scenario.html#manifest_module?/1"},{"type":"function","doc":"","title":"Runbox.Scenario.template_module?/2","ref":"Runbox.Scenario.html#template_module?/2"},{"type":"type","doc":"","title":"Runbox.Scenario.component_def/0","ref":"Runbox.Scenario.html#t:component_def/0"},{"type":"type","doc":"","title":"Runbox.Scenario.t/0","ref":"Runbox.Scenario.html#t:t/0"},{"type":"module","doc":"Module contains functions for working with scenario notifications.","title":"Runbox.Scenario.Notification","ref":"Runbox.Scenario.Notification.html"},{"type":"function","doc":"Lists all channels with templates defined for scenario, notification type and template type.","title":"Runbox.Scenario.Notification.list_channels/3","ref":"Runbox.Scenario.Notification.html#list_channels/3"},{"type":"function","doc":"Lists all template languages defined for scenario, notification type, template type and channel.","title":"Runbox.Scenario.Notification.list_languages/4","ref":"Runbox.Scenario.Notification.html#list_languages/4"},{"type":"function","doc":"Lists all notification types for a scenario.","title":"Runbox.Scenario.Notification.list_notification_types/1","ref":"Runbox.Scenario.Notification.html#list_notification_types/1"},{"type":"function","doc":"Lists all template types of the specified notification.","title":"Runbox.Scenario.Notification.list_template_types/2","ref":"Runbox.Scenario.Notification.html#list_template_types/2"},{"type":"function","doc":"Loads and evaluates templates specified by scenario, notification type, template type, channel\nand language.\n\nIf language is not defined or does not exist, `fallback_language` is used.","title":"Runbox.Scenario.Notification.load_and_eval_templates/7","ref":"Runbox.Scenario.Notification.html#load_and_eval_templates/7"},{"type":"module","doc":"Template inspector for Simple scenario.\n\nValidates and provides basic information about a Simple scenario template.","title":"Runbox.Scenario.Template.Simple","ref":"Runbox.Scenario.Template.Simple.html"},{"type":"behaviour","doc":"Behaviour used to detect scenario template source files.","title":"Runbox.Scenario.TemplateInspector","ref":"Runbox.Scenario.TemplateInspector.html"},{"type":"callback","doc":"","title":"Runbox.Scenario.TemplateInspector.template_info/1","ref":"Runbox.Scenario.TemplateInspector.html#c:template_info/1"},{"type":"function","doc":"Retrieve implementation specific information about a template.","title":"Runbox.Scenario.TemplateInspector.template_info/2","ref":"Runbox.Scenario.TemplateInspector.html#template_info/2"},{"type":"callback","doc":"","title":"Runbox.Scenario.TemplateInspector.valid_template?/1","ref":"Runbox.Scenario.TemplateInspector.html#c:valid_template?/1"},{"type":"function","doc":"Validate whether given module is valid template source file.","title":"Runbox.Scenario.TemplateInspector.valid_template?/2","ref":"Runbox.Scenario.TemplateInspector.html#valid_template?/2"},{"type":"module","doc":"Module for inspecting a scenario release.","title":"Runbox.ScenarioRelease","ref":"Runbox.ScenarioRelease.html"},{"type":"function","doc":"Retrieves information about scenarios in a release.","title":"Runbox.ScenarioRelease.release_info/1","ref":"Runbox.ScenarioRelease.html#release_info/1"},{"type":"module","doc":"Functions which are called on the scenario slave via RPC from the master\nnode.","title":"Runbox.ScenarioRelease.SlaveFunc","ref":"Runbox.ScenarioRelease.SlaveFunc.html"},{"type":"function","doc":"","title":"Runbox.ScenarioRelease.SlaveFunc.release_info/1","ref":"Runbox.ScenarioRelease.SlaveFunc.html#release_info/1"},{"type":"function","doc":"Returns all modules of the scenario app.","title":"Runbox.ScenarioRelease.SlaveFunc.release_modules/0","ref":"Runbox.ScenarioRelease.SlaveFunc.html#release_modules/0"},{"type":"function","doc":"Returns list of scenarios in release.","title":"Runbox.ScenarioRelease.SlaveFunc.scenarios/0","ref":"Runbox.ScenarioRelease.SlaveFunc.html#scenarios/0"},{"type":"module","doc":"Holds information about scenario template.","title":"Runbox.ScenarioTemplate","ref":"Runbox.ScenarioTemplate.html"},{"type":"type","doc":"","title":"Runbox.ScenarioTemplate.t/0","ref":"Runbox.ScenarioTemplate.html#t:t/0"},{"type":"module","doc":"Utilities for starting a slave node with a scenario release and communicating\nwith the node.","title":"Runbox.Slave","ref":"Runbox.Slave.html"},{"type":"function","doc":"Calls a function on the slave.\n\nCalls MFA on the `slave`. See `:erpc.call/4` for details.","title":"Runbox.Slave.call/4","ref":"Runbox.Slave.html#call/4"},{"type":"function","doc":"Starts a slave node.\n\nThe `release_dir` is a directory with scenario release.\n\nIf `app_env` is set, given app environment variables are set on slave.","title":"Runbox.Slave.start/2","ref":"Runbox.Slave.html#start/2"},{"type":"function","doc":"Spawns a slave and executes the `fun`.\n\n`fun` receives the slave node as a parameter. The slave is terminated after\nthe `fun` finishes.","title":"Runbox.Slave.with_slave/2","ref":"Runbox.Slave.html#with_slave/2"},{"type":"module","doc":"Module defines struct which acts as state store envelope for any stateful runtime component.","title":"Runbox.StateStore.Entity","ref":"Runbox.StateStore.Entity.html"},{"type":"function","doc":"Confirm that all messages with lower timestamps were processed.\n\nWhen the timestamp is higher than next savepoint the state is persisted into `StateStore` with\n`StateStore.save/4` and next savepoint target is determined and stored in the entity.\nThe state is persisted only once per every savepoint.\n\nOptionally you can specify an entity transformation that is applied before save. This\ntransformation is only execute before save, so it can be expensive. The transformation returns\neither `{:ok, entity}` or `{:error, any()}` if there's any error during the transformation\npreventing the entity to be saved.","title":"Runbox.StateStore.Entity.ack_processed_time/4","ref":"Runbox.StateStore.Entity.html#ack_processed_time/4"},{"type":"function","doc":"Returns entity id","title":"Runbox.StateStore.Entity.id/1","ref":"Runbox.StateStore.Entity.html#id/1"},{"type":"function","doc":"Creates new state store entity","title":"Runbox.StateStore.Entity.new/5","ref":"Runbox.StateStore.Entity.html#new/5"},{"type":"function","doc":"Returns entity schedule","title":"Runbox.StateStore.Entity.schedule/1","ref":"Runbox.StateStore.Entity.html#schedule/1"},{"type":"function","doc":"Returns entity state","title":"Runbox.StateStore.Entity.state/1","ref":"Runbox.StateStore.Entity.html#state/1"},{"type":"function","doc":"Returns entity timestamp","title":"Runbox.StateStore.Entity.timestamp/1","ref":"Runbox.StateStore.Entity.html#timestamp/1"},{"type":"function","doc":"Set entity `state` at the specified point in time.","title":"Runbox.StateStore.Entity.update_state/3","ref":"Runbox.StateStore.Entity.html#update_state/3"},{"type":"type","doc":"","title":"Runbox.StateStore.Entity.id/0","ref":"Runbox.StateStore.Entity.html#t:id/0"},{"type":"type","doc":"","title":"Runbox.StateStore.Entity.state/0","ref":"Runbox.StateStore.Entity.html#t:state/0"},{"type":"type","doc":"","title":"Runbox.StateStore.Entity.t/0","ref":"Runbox.StateStore.Entity.html#t:t/0"},{"type":"type","doc":"","title":"Runbox.StateStore.Entity.t/1","ref":"Runbox.StateStore.Entity.html#t:t/1"},{"type":"module","doc":"Module contains functions to compute savepoint timestamps.\n\nSavepoint timestamps are calculated as multiples of `schedule`.","title":"Runbox.StateStore.ScheduleUtils","ref":"Runbox.StateStore.ScheduleUtils.html"},{"type":"function","doc":"Returns next savepoint based on given `schedule` and `timestamp`.\n\nNext savepoint is calculated as next multiple of `schedule` larger or equal to given\n`timestamp`.","title":"Runbox.StateStore.ScheduleUtils.next_savepoint/2","ref":"Runbox.StateStore.ScheduleUtils.html#next_savepoint/2"},{"type":"function","doc":"```\niex> Runbox.StateStore.ScheduleUtils.next_savepoint(60_000, 0)\n60_000\n\niex> Runbox.StateStore.ScheduleUtils.next_savepoint(60_000, 10_000)\n60_000\n\niex> Runbox.StateStore.ScheduleUtils.next_savepoint(60_000, 60_000)\n120_000\n```","title":"Examples - Runbox.StateStore.ScheduleUtils.next_savepoint/2","ref":"Runbox.StateStore.ScheduleUtils.html#next_savepoint/2-examples"},{"type":"function","doc":"Returns previous savepoint based on given `schedule` and `timestamp`.\n\nNext savepoint is calculated as previous multiple of `schedule` lower or equal to given\n`timestamp`.","title":"Runbox.StateStore.ScheduleUtils.previous_savepoint/2","ref":"Runbox.StateStore.ScheduleUtils.html#previous_savepoint/2"},{"type":"function","doc":"```\niex> Runbox.StateStore.ScheduleUtils.previous_savepoint(60_000, 30_000)\n0\n\niex> Runbox.StateStore.ScheduleUtils.previous_savepoint(60_000, 60_000)\n60_000\n\niex> Runbox.StateStore.ScheduleUtils.previous_savepoint(60_000, 90_000)\n60_000\n\niex> Runbox.StateStore.ScheduleUtils.previous_savepoint(60_000, 120_000)\n120_000\n\niex> Runbox.StateStore.ScheduleUtils.previous_savepoint(60_000, 150_000)\n120_000\n```","title":"Examples - Runbox.StateStore.ScheduleUtils.previous_savepoint/2","ref":"Runbox.StateStore.ScheduleUtils.html#previous_savepoint/2-examples"},{"type":"function","doc":"Returns unreached savepoints based on given `schedule` in range of given timestamps `(from; to>`","title":"Runbox.StateStore.ScheduleUtils.unreached_savepoints/3","ref":"Runbox.StateStore.ScheduleUtils.html#unreached_savepoints/3"},{"type":"function","doc":"```\niex> Runbox.StateStore.ScheduleUtils.unreached_savepoints(60_000, 0, 120_000)\n[60_000, 120_000]\n\niex> Runbox.StateStore.ScheduleUtils.unreached_savepoints(60_000, 60_000, 120_000)\n[120_000]\n\niex> Runbox.StateStore.ScheduleUtils.unreached_savepoints(60_000, 60_000, 180_000)\n[120_000, 180_000]\n```","title":"Examples - Runbox.StateStore.ScheduleUtils.unreached_savepoints/3","ref":"Runbox.StateStore.ScheduleUtils.html#unreached_savepoints/3-examples"},{"type":"function","doc":"Returns count of unreached savepoints.","title":"Runbox.StateStore.ScheduleUtils.unreached_savepoints_count/3","ref":"Runbox.StateStore.ScheduleUtils.html#unreached_savepoints_count/3"},{"type":"function","doc":"```\niex> Runbox.StateStore.ScheduleUtils.unreached_savepoints_count(60_000, 0, 120_000)\n2\n\niex> Runbox.StateStore.ScheduleUtils.unreached_savepoints_count(60_000, 60_000, 120_000)\n1\n\niex> Runbox.StateStore.ScheduleUtils.unreached_savepoints_count(1, 0, 120_000)\n120_000\n```","title":"Examples - Runbox.StateStore.ScheduleUtils.unreached_savepoints_count/3","ref":"Runbox.StateStore.ScheduleUtils.html#unreached_savepoints_count/3-examples"},{"type":"type","doc":"Timestamp defined as unix epoch in milliseconds","title":"Runbox.StateStore.ScheduleUtils.epoch_ms/0","ref":"Runbox.StateStore.ScheduleUtils.html#t:epoch_ms/0"},{"type":"type","doc":"Difference between two unix epoch timestamps in milliseconds","title":"Runbox.StateStore.ScheduleUtils.epoch_ms_interval/0","ref":"Runbox.StateStore.ScheduleUtils.html#t:epoch_ms_interval/0"},{"type":"type","doc":"Interval (in seconds) between savepoint timestamps.","title":"Runbox.StateStore.ScheduleUtils.schedule/0","ref":"Runbox.StateStore.ScheduleUtils.html#t:schedule/0"},{"type":"module","doc":"Dead-letter message produced when evaluating of raw-topic failed.","title":"Runbox.ErrorMessage","ref":"Runbox.ErrorMessage.html"},{"type":"type","doc":"","title":"Runbox.ErrorMessage.t/0","ref":"Runbox.ErrorMessage.html#t:t/0"},{"type":"module","doc":"Formerly delayed messages/late events, when jitter around incoming topic is out of threshold.\n\nNegative delay means message is too late (from past). Positive delay means message is too early\n(from future).\n\nNormalized messages is provided.","title":"Runbox.JitterMessage","ref":"Runbox.JitterMessage.html"},{"type":"type","doc":"See module `Runbox.JitterMessage` info.","title":"Runbox.JitterMessage.t/0","ref":"Runbox.JitterMessage.html#t:t/0"},{"type":"module","doc":"Altworx message runtime representation.","title":"Runbox.Message","ref":"Runbox.Message.html"},{"type":"function","doc":"","title":"Runbox.Message.body/1","ref":"Runbox.Message.html#body/1"},{"type":"function","doc":"","title":"Runbox.Message.from/1","ref":"Runbox.Message.html#from/1"},{"type":"function","doc":"","title":"Runbox.Message.new/4","ref":"Runbox.Message.html#new/4"},{"type":"function","doc":"","title":"Runbox.Message.timestamp/1","ref":"Runbox.Message.html#timestamp/1"},{"type":"function","doc":"","title":"Runbox.Message.type/1","ref":"Runbox.Message.html#type/1"},{"type":"type","doc":"","title":"Runbox.Message.body/0","ref":"Runbox.Message.html#t:body/0"},{"type":"type","doc":"","title":"Runbox.Message.from/0","ref":"Runbox.Message.html#t:from/0"},{"type":"type","doc":"","title":"Runbox.Message.origin/0","ref":"Runbox.Message.html#t:origin/0"},{"type":"type","doc":"","title":"Runbox.Message.t/0","ref":"Runbox.Message.html#t:t/0"},{"type":"type","doc":"","title":"Runbox.Message.timestamp/0","ref":"Runbox.Message.html#t:timestamp/0"},{"type":"type","doc":"","title":"Runbox.Message.type/0","ref":"Runbox.Message.html#t:type/0"},{"type":"module","doc":"Struct produced by scenario representing instruction for runtime.","title":"Runbox.Runtime.RuntimeInstruction","ref":"Runbox.Runtime.RuntimeInstruction.html"},{"type":"function","doc":"Creates an runtime instruction to register a given timeout message.\n\nTimeouts allow scenarios to be notified at a specified time. This is useful when you want to\nperform something for example in 10 minutes from now or at tomorrow's noon (both in terms\nof the external time).\n\nWhen registering a timeout, you have to provide a message. This message will arrive to your\nrun exactly at the time specified inside its `timestamp` field. Which component of your run\nwill handle the message depends on the type of runtime that your scenario is based on.\nSee docs of the particular runtime type for more details.\n\nUsually, you can use the return value of `register_timeout/1` wherever an output action is\nexpected.","title":"Runbox.Runtime.RuntimeInstruction.register_timeout/1","ref":"Runbox.Runtime.RuntimeInstruction.html#register_timeout/1"},{"type":"function","doc":"register_timeout(%Runbox.Message{\n      timestamp: now + :timer.minutes(10),\n      type: :check_if_asset_has_been_repaired_and_send_a_notification_if_not,\n      body: %{\n        broken_asset: %{type: \"/assets/router\", id: \"192.168.100.115\"}\n      }\n    })","title":"Example - Runbox.Runtime.RuntimeInstruction.register_timeout/1","ref":"Runbox.Runtime.RuntimeInstruction.html#register_timeout/1-example"},{"type":"type","doc":"Struct produced by scenario representing instruction for runtime.","title":"Runbox.Runtime.RuntimeInstruction.t/0","ref":"Runbox.Runtime.RuntimeInstruction.html#t:t/0"},{"type":"module","doc":"Instruction to register unit timeout.\n\nUse `Runbox.Runtime.RuntimeInstruction.register_timeout/1` to create this runtime instruction.","title":"Runbox.Runtime.RuntimeInstruction.Timeout","ref":"Runbox.Runtime.RuntimeInstruction.Timeout.html"},{"type":"type","doc":"Struct representing unit timeout registration.","title":"Runbox.Runtime.RuntimeInstruction.Timeout.t/0","ref":"Runbox.Runtime.RuntimeInstruction.Timeout.html#t:t/0"},{"type":"module","doc":"Output action represents a side effect produced by a run.","title":"Runbox.Scenario.OutputAction","ref":"Runbox.Scenario.OutputAction.html"},{"type":"module","doc":"Creating output action typically happens from scenario. It is done by\ncreating one of the appropriate struct:\n\n  * `t:Runbox.Scenario.OutputAction.UpsertAssetAttributes.t/0`\n  * `t:Runbox.Scenario.OutputAction.DeleteAssetAttributes.t/0`\n  * `t:Runbox.Scenario.OutputAction.DeleteAllAssetAttributes.t/0`\n  * `t:Runbox.Scenario.OutputAction.UpsertEdge.t/0`\n  * `t:Runbox.Scenario.OutputAction.DeleteEdge.t/0`\n  * `t:Runbox.Scenario.OutputAction.Notification.t/0`\n  * `t:Runbox.Scenario.OutputAction.Event.t/0`\n  * `t:Runbox.Scenario.OutputAction.ExecuteSQL.t/0`\n  * `t:Runbox.Scenario.OutputAction.Incident.t/0`\n  * `t:Runbox.Scenario.OutputAction.IncidentPatch.t/0`\n\nFor example:\n\n    iex> %OutputAction.UpsertAssetAttributes{\n    ...>   type: \"/asset/camera\",\n    ...>   id: \"one\",\n    ...>   attributes: %{\"foo\" => \"bar\"}\n    ...> }","title":"Creating output actions - Runbox.Scenario.OutputAction","ref":"Runbox.Scenario.OutputAction.html#module-creating-output-actions"},{"type":"module","doc":"Changes done by output actions to assets and edges are scoped for each\nscenario run. The global result is computed from the changes done by\nindividual runs.\n\nA single run is forbidden to do multiple changes to the same object (asset\nattribute or edge) at the same time. However distinct runs can perform such\nchanges, because they are scoped to each run.\n\nThe rules to compute the global state apply in the following order:\n\n  1. Change with the higher timestamp has precedence.\n\n  1. Updating an asset attribute or an edge has precedence to deleting it.\n\n  1. Change with the higher value has precedence (applies only for asset attributes).","title":"Asset and edge output actions - Runbox.Scenario.OutputAction","ref":"Runbox.Scenario.OutputAction.html#module-asset-and-edge-output-actions"},{"type":"type","doc":"","title":"Runbox.Scenario.OutputAction.access_tag/0","ref":"Runbox.Scenario.OutputAction.html#t:access_tag/0"},{"type":"type","doc":"Asset ID.\n\nAsset ID together with an asset type uniquely identifies an asset.\n\nAsset ID must not contain any slashes. They can contain any other character though.\nFor example, these are valid asset IDs: `fire_alarm`, `192.168.0.142`, `Meeting room`, and\n`Příliš žluťoučký kůň`.","title":"Runbox.Scenario.OutputAction.asset_id/0","ref":"Runbox.Scenario.OutputAction.html#t:asset_id/0"},{"type":"type","doc":"Asset type.\n\nAssets are grouped by their asset types. Assets of the same type should be of similar nature\n(e.g. servers or buildings) and should share the same set of attributes.\n\nAsset types must start with a slash and must not end with a slash. Slashes can be used\nwithin the asset type to denote structure. For example, these are valid asset types:\n`/assets/servers`, `/servers`, and `/assets/network/routers`.\n\nAlthough asset types like `/servers` are permitted, it is recommended to prefix asset types\nwith the `/assets` prefix to enable consistent configuration of common attributes (e.g., `name`)\nvia the root `/assets` asset type in the UI.","title":"Runbox.Scenario.OutputAction.asset_type/0","ref":"Runbox.Scenario.OutputAction.html#t:asset_type/0"},{"type":"type","doc":"","title":"Runbox.Scenario.OutputAction.edge_type/0","ref":"Runbox.Scenario.OutputAction.html#t:edge_type/0"},{"type":"type","doc":"String with interpolation support.\n\nSupports interpolating asset references to a format which is useful for\ndisplaying assets in the Altworx UI. The string can contain any number of\nplaceholders in format\n\n    ${assets.[\"/assets/camera\", \"one\"]}\n\nWhere `\"/assets/camera\"` is the asset type and `\"one\"` is the asset ID. The resulting interpolation is\n\n    [Camera One[/assets/camera/one]]\n\nWhere `Camera One` is the asset's `name` attribute and `/assets/camera/one`\nis the asset's full ID. When the asset `name` is not present, the asset's ID\nis used instead\n\n    [one[/assets/camera/one]]\n\nWhen the asset cannot be found, the resulting interpolation is\n\n    unknown\n\nThe interpolation is done upon \"read\" operations, therefore the result can\nvary based on the asset visibility for the identity of the user performing\nthe operation.","title":"Runbox.Scenario.OutputAction.interpolable/0","ref":"Runbox.Scenario.OutputAction.html#t:interpolable/0"},{"type":"type","doc":"","title":"Runbox.Scenario.OutputAction.oa_params/0","ref":"Runbox.Scenario.OutputAction.html#t:oa_params/0"},{"type":"type","doc":"Output Action\n\nThis struct is produced by `Runbox.Runtime.Stage.Sandbox.execute_run/3`.","title":"Runbox.Scenario.OutputAction.t/0","ref":"Runbox.Scenario.OutputAction.html#t:t/0"},{"type":"module","doc":"Toolbox for working with scenario user actions.\n\nUser action is action created by scenario via `Runbox.Scenario.UIAction` and fired\nby user (usually from UI). Scenario is usually also a receiver of fired action.\nAssumed workflow is following:\n\n  1) Scenario creates list of actions (using `Runbox.Scenario.UIAction`) to be fired\n     in the future and saves it somewhere (in incident asset for example).\n\n  2) UI displays available actions.\n\n  3) User fires selected action from UI.\n\n  4) UI pushes the action to raw Kafka topic (defined in action) with extra data attached\n     (usually some comment).\n\n  5) Normalizer pipeline receives the action and sends it to target scenario.","title":"Runbox.Scenario.UserAction","ref":"Runbox.Scenario.UserAction.html"},{"type":"function","doc":"Unpacks action and makes it fired (prepared for sending to target topic).\n\nRouting properties (`type`, `details`) are propagated from packed action, `user` and `data`\nproperties are attached to fired action. Target `topic` is also extracted from packed action.\n\nReturns error if packed action (JWT) is not valid.","title":"Runbox.Scenario.UserAction.fire/3","ref":"Runbox.Scenario.UserAction.html#fire/3"},{"type":"function","doc":"Helper for UserAction forwarding to be used in normalizer.\n\nIf helper is used in normalizer pipeline configuration in `config.ini` like\n\n    [pipelines.some_pipeline]\n    definition = Runbox.Scenario.UserAction.normalizer_pipeline_step\n\nthen `Runbox.Message` `body` will contain fired `Runbox.Scenario.UserAction`.","title":"Runbox.Scenario.UserAction.normalizer_pipeline_step/1","ref":"Runbox.Scenario.UserAction.html#normalizer_pipeline_step/1"},{"type":"function","doc":"Packs user action to JWT to be used (fired) later (from UI for example).\n\nParameter `details` must be a map encodable to JSON. Because `details` can be (and it is!)\nencoded/decoded to/from JSON during its life-cycle, atom property names of `details` are\nalways converted to string property names in fired action's `details`.\n\nJWT is not used for security reasons - it is just good enough envelope protecting\naction against accidental corruption on its journey.","title":"Runbox.Scenario.UserAction.pack/3","ref":"Runbox.Scenario.UserAction.html#pack/3"},{"type":"function","doc":"Parses string containing JSON to `Runbox.Scenario.UserAction`.","title":"Runbox.Scenario.UserAction.parse_raw_action/1","ref":"Runbox.Scenario.UserAction.html#parse_raw_action/1"},{"type":"type","doc":"Fired user action.\n\nWhen action is fired by the user, it should be pushed to raw `topic` with current\ntimestamp to be correctly reordered by `Normalizer`. Properties `type` and `details`\nshould contain enough information for routing action to the target scenario. Property\n`user` is ID of user who fired the action, `data` contains extra information attached\nby environment where user fired the action (for example comment filled by the user on UI).","title":"Runbox.Scenario.UserAction.t/0","ref":"Runbox.Scenario.UserAction.html#t:t/0"},{"type":"module","doc":"Body of invalid output action.","title":"Runbox.Scenario.OutputAction.BadOutputAction","ref":"Runbox.Scenario.OutputAction.BadOutputAction.html"},{"type":"type","doc":"Bad Output Action\n\n  * `:params` - holds the original parameters","title":"Runbox.Scenario.OutputAction.BadOutputAction.t/0","ref":"Runbox.Scenario.OutputAction.BadOutputAction.html#t:t/0"},{"type":"module","doc":"Parameters for output action Delete All Asset Attributes.\n\nThe resulting output action will delete all existing attributes of the\ngiven asset. This is scoped to the attributes created by the same run.","title":"Runbox.Scenario.OutputAction.DeleteAllAssetAttributes","ref":"Runbox.Scenario.OutputAction.DeleteAllAssetAttributes.html"},{"type":"type","doc":"Delete All Asset Attributes\n\n  * `:type` - asset type\n  * `:id` - asset ID","title":"Runbox.Scenario.OutputAction.DeleteAllAssetAttributes.t/0","ref":"Runbox.Scenario.OutputAction.DeleteAllAssetAttributes.html#t:t/0"},{"type":"module","doc":"Parameters for output action Delete Asset Attributes.\n\nThe resulting output action will delete attributes of the given asset. This\naction succeeds even if the attributes do not exist.","title":"Runbox.Scenario.OutputAction.DeleteAssetAttributes","ref":"Runbox.Scenario.OutputAction.DeleteAssetAttributes.html"},{"type":"type","doc":"Delete Asset Attributes\n\n  * `:type` - asset type\n  * `:id` - asset ID\n  * `:attributes` - a map of attributes to be deleted\n\nThe leaf attributes of the `:attributes` map are deleted.\nE.g., if asset has the following attributes\n```elixir\n%{\n  \"a\": 1,\n  \"b\": %{\"c\" => 2}\n}\n```\nand the following `:attributes` are provided in the output action\n```elixir\n%{\n  \"b\": %{\"c\" => true}\n}\n```\nthen `b.c` attribute will be deleted, resulting in\n```elixir\n%{\n  \"a\": 1\n  \"b\": %{}\n}\n```","title":"Runbox.Scenario.OutputAction.DeleteAssetAttributes.t/0","ref":"Runbox.Scenario.OutputAction.DeleteAssetAttributes.html#t:t/0"},{"type":"module","doc":"Parameters for output action Delete Edge.\n\nThe resulting output action will delete an edge between two assets. This\naction succeeds even if the edge does not exist.","title":"Runbox.Scenario.OutputAction.DeleteEdge","ref":"Runbox.Scenario.OutputAction.DeleteEdge.html"},{"type":"type","doc":"Delete Edge\n\n  * `:from_type` - type of the asset which this edge goes out from\n  * `:from_id` - id of the asset which this edge goes out from\n  * `:to_type` - type of the asset which this edge goes to\n  * `:to_id` - id of the asset which this edge goes to\n  * `:type` - type of the edge","title":"Runbox.Scenario.OutputAction.DeleteEdge.t/0","ref":"Runbox.Scenario.OutputAction.DeleteEdge.html#t:t/0"},{"type":"module","doc":"Parameters for output action Event.\n\nThe resulting output action creates an event.\n\nSee `t:t/0` for more information.\n\n> #### Note {: .warning}\n>\n> Beware that creating incident history manually via event's `incident_actors` property, instead\n> of relying on Incident output actions, is low-level and prone to errors. You must always ensure\n> that you create the event **and** you accordingly update the incident itself. Failing to do\n> either will result in inconsistent state.","title":"Runbox.Scenario.OutputAction.Event","ref":"Runbox.Scenario.OutputAction.Event.html"},{"type":"type","doc":"","title":"Runbox.Scenario.OutputAction.Event.actor/0","ref":"Runbox.Scenario.OutputAction.Event.html#t:actor/0"},{"type":"type","doc":"","title":"Runbox.Scenario.OutputAction.Event.actors/0","ref":"Runbox.Scenario.OutputAction.Event.html#t:actors/0"},{"type":"type","doc":"Incident actor of the event.\n\nCan mean two things depending on if `status` and `severity` is set. If either is empty, then the\nevent is considered related to the incident, similar to `actors`. If both are set, then the\nevent is considered a part of the incident's history. `status` and `severity` then represent the\nchange of these incident properties at the time of the event. `attributes` is fully optional\nand contains further context about the change of the incident.","title":"Runbox.Scenario.OutputAction.Event.incident_actor/0","ref":"Runbox.Scenario.OutputAction.Event.html#t:incident_actor/0"},{"type":"type","doc":"","title":"Runbox.Scenario.OutputAction.Event.incident_actors/0","ref":"Runbox.Scenario.OutputAction.Event.html#t:incident_actors/0"},{"type":"type","doc":"Event\n\n  * `:type` - type of the event. It's declared in Scenario's Manifest.\n  * `:template` - event's template that allows to interpolate actors and parameters\n    (see the *Interpolation* section below).\n  * `:actors` - map of asset actors that are to be linked with the event. They may also be\n    interpolated in the template.\n  * `:incident_actors` - map of incident actors that are to be linked with the event. This means\n    the event is either part of the incident's history, or is just related. They may also be\n    interpolated in the template. See `t:incident_actor/0` for more information.\n  * `:params` - map of template parameters that are interpolated in the template\n    (optional, defaults to `%{}`).\n  * `:origin_messages` - list of references to raw messages linked with the event itself\n    (optional, defaults to `[]`).","title":"Runbox.Scenario.OutputAction.Event.t/0","ref":"Runbox.Scenario.OutputAction.Event.html#t:t/0"},{"type":"type","doc":"The template may reference actors using placeholders like `${actors.actor_key}`, where\n`actor_key` corresponds to a key in the `actors` map within this struct. In the UI,\nthese placeholders are transformed into links. Each link points to the corresponding\nasset and displays the asset's name as the link text.\n\nIncident actors can be interpolated in a very similar manner using `incidents` keyword -\n`${incidents.fire}`.\n\nThe template may also contain placeholders like `${params.param_key}`, where `param_key`\nis a key under the `params` map. These placeholders are simply replaced with the\ncorresponding values from the `params` map. These parameters are useful for dynamic\ncontent other than actors.","title":"Interpolation - Runbox.Scenario.OutputAction.Event.t/0","ref":"Runbox.Scenario.OutputAction.Event.html#t:t/0-interpolation"},{"type":"type","doc":"%Event{\n      type: \"server_login\",\n      template: \"${actors.person} logged into ${actors.server} using OpenSSH ${params.openssh_version}\",\n      actors: %{\n        \"person\" => %{\n          asset_type: \"/assets/person\",\n          asset_id: \"joe\"\n        },\n        \"server\" => %{\n          asset_type: \"/assets/server\",\n          asset_id: \"192.168.142.18\"\n        }\n      },\n      params: %{\"openssh_version\" => \"9.8\"},\n      origin_messages: [normalized_message.origin]\n    }","title":"Example - Runbox.Scenario.OutputAction.Event.t/0","ref":"Runbox.Scenario.OutputAction.Event.html#t:t/0-example"},{"type":"module","doc":"Parameters for output action Execute SQL.\n\nThe resulting output action executes an SQL statement.","title":"Runbox.Scenario.OutputAction.ExecuteSQL","ref":"Runbox.Scenario.OutputAction.ExecuteSQL.html"},{"type":"type","doc":"Execute SQL\n\n  * `:db_connection` - map with database connection info (`hostname`, `username`, `password`, `database`)\n  * `:sql_query` - SQL query.\n  * `:data` - list of parameterized data.\n  * `:type` - database type, right now only PostgreSQL is supported. It's checked in `OutputProcessor` and\n    the value must be `:postgresql`.\n\nIf `:run_id` is present in `data`, it's replaced with actual run_id in `AssetMap.Api.SqlExecutor`.","title":"Runbox.Scenario.OutputAction.ExecuteSQL.t/0","ref":"Runbox.Scenario.OutputAction.ExecuteSQL.html#t:t/0"},{"type":"module","doc":"Parameters for output action Incident.\n\nThe resulting output action creates an incident.","title":"Runbox.Scenario.OutputAction.Incident","ref":"Runbox.Scenario.OutputAction.Incident.html"},{"type":"type","doc":"","title":"Runbox.Scenario.OutputAction.Incident.severity/0","ref":"Runbox.Scenario.OutputAction.Incident.html#t:severity/0"},{"type":"type","doc":"Incident","title":"Runbox.Scenario.OutputAction.Incident.t/0","ref":"Runbox.Scenario.OutputAction.Incident.html#t:t/0"},{"type":"module","doc":"Incident Actor","title":"Runbox.Scenario.OutputAction.IncidentActor","ref":"Runbox.Scenario.OutputAction.IncidentActor.html"},{"type":"type","doc":"","title":"Runbox.Scenario.OutputAction.IncidentActor.t/0","ref":"Runbox.Scenario.OutputAction.IncidentActor.html#t:t/0"},{"type":"module","doc":"Incident Future.\n\nThe struct represents a future possible change of an Incident. It carries information about how\nthe Incident might change.\n\nThe change can be expected to happen at a specified time (then `timestamp` is a positive integer\n- epoch time in ms) or it is unknown when the change might occur (then `timestamp` is `nil`).","title":"Runbox.Scenario.OutputAction.IncidentFuture","ref":"Runbox.Scenario.OutputAction.IncidentFuture.html"},{"type":"type","doc":"","title":"Runbox.Scenario.OutputAction.IncidentFuture.t/0","ref":"Runbox.Scenario.OutputAction.IncidentFuture.html#t:t/0"},{"type":"module","doc":"Incident history.\n\nThe struct represents a change of an Incident in time. It carries information about how the\nIncident changed, all describe how the Incident looks after the change. The struct contains all\nnecessary parameters to create an Incident history record.\n\nSince Incident history is now synonymous to Events, `event_*` fields of this struct can be used\nto control how the final Event looks like.\n\nAvailable fields\n\n- `:status` - status of the Incident\n- `:severity` - severity of the Incident\n- `:timestamp` - time of the change\n- `:description` - description of the change - what has happened, will be used as Event template\n- `:attributes` - (optional) significant additional attributes that changed with the Incident\n  (usually metrics, like temperature, server load, distance). Can be empty.\n- `:event_type` - (optional) type of the underlying Event, if empty a default type\n  `incident_updated` will be used\n- `:event_actors` - (optional) additional actors of the underlying Event. If empty no actors are\n  defined.\n- `:event_params` - (optional) additional parameters of the Event. If empty no params are\n  defined.\n- `:event_origin_messages` - (optional) list of references to raw messages linked with the\n  underlying Event. If empty no raw messages are linked with the Event.","title":"Runbox.Scenario.OutputAction.IncidentHistory","ref":"Runbox.Scenario.OutputAction.IncidentHistory.html"},{"type":"type","doc":"Incident history.\n\nNote Incident history is synonymous to Events and thus carries similar parameters.\n\n- `:status` - status of the Incident\n- `:severity` - severity of the Incident\n- `:timestamp` - time of the change\n- `:description` - description of the change - what has happened, will be used as Event template\n- `:attributes` - (optional) significant additional attributes that changed with the Incident\n  (usually metrics, like temperature, server load, distance). Can be empty.\n- `:event_type` - (optional) type of the underlying Event, if empty a default type\n  `incident_updated` will be used\n- `:event_actors` - (optional) additional actors of the underlying Event. If empty no actors are\n  defined.\n- `:event_params` - (optional) additional parameters of the Event. If empty no params are\n  defined.\n- `:event_origin_messages` - (optional) list of references to raw messages linked with the\n  underlying Event. If empty no raw messages are linked with the Event.","title":"Runbox.Scenario.OutputAction.IncidentHistory.t/0","ref":"Runbox.Scenario.OutputAction.IncidentHistory.html#t:t/0"},{"type":"module","doc":"Parameters for output action Incident Patch.\n\nThe resulting output action updates an incident.","title":"Runbox.Scenario.OutputAction.IncidentPatch","ref":"Runbox.Scenario.OutputAction.IncidentPatch.html"},{"type":"type","doc":"Incident Patch","title":"Runbox.Scenario.OutputAction.IncidentPatch.t/0","ref":"Runbox.Scenario.OutputAction.IncidentPatch.html#t:t/0"},{"type":"module","doc":"Parameters for output action Notification.\n\nThe resulting output action sends a notification.","title":"Runbox.Scenario.OutputAction.Notification","ref":"Runbox.Scenario.OutputAction.Notification.html"},{"type":"type","doc":"Notification\n\n  * `:type` - string, type of the notification\n  * `:data` - map, data for the template evaluation\n  * `:primary_asset` - string, optional, id of an asset that corresponds to this notification\n  * `:ids` - list of asset ids, optional, to filter recipients, each recipient must have\n    sufficient privileges to see all listed IDs\n  * `:aqls` - map, AQLs which are evaluated, provide further context to the templates, and also can\n    serve the same purpose as `ids` - filter recipients to only those who can see the whole\n    result of the query. Key is the name of the AQL which can be used in the template. Value\n    is a map with `aql` (the AQL query) and `primary` (whether the AQL should also be used\n    to filter recipients) keys.\n  * `:priority` - optional, defaults to `:medium`, can be `:low`, `:medium`, `:high`, only\n    applicable to some channels\n  * `:metadata` - map, optional, additional metadata which can be used in notification group\n    subscriptions. Should adhere to the format specified in `Runbox.Scenario.Manifest.notifications`.\n  * `:direct_subscriptions` - list of maps, subscriptions to be used additionally to the usual\n    routing. Each map should have the following.\n      * `:user` - map, user object to be used instead of Altworx object, should have\n        `notification.email` and optionally even `notification.language`.\n      * `:templates` - list of ids, which templates should be used\n      * `:channels` - list of ids, which channels should be used\n  * `:attachments` - list of maps, optional, attachments that should be sent with the notification.\n    Each map has the following keys.\n      * `:data` - binary, content of the attachment\n      * `:filename` - string, name of the file\n      * `:content_type` - string, optional, MIME type of the data\n      * `:type` - whether the attachment is to inlined in the template (`:inline`) or not\n        (`:attachment`, this is the default)\n  * `:email_reply_to` - map, optional, used to specify `reply_to` parameter of the notification email\n    when it is sent using the email notification channel. If the field si left nil, default\n    `reply_to` parameter from configuration is used.\n      * `key: string` - name of the person the reader should reply to\n      * `value: string` - email of the person the reader should reply to","title":"Runbox.Scenario.OutputAction.Notification.t/0","ref":"Runbox.Scenario.OutputAction.Notification.html#t:t/0"},{"type":"module","doc":"Parameters for output action Upsert Asset Attributes.\n\nThe resulting output action will create or update attributes of the given\nasset.","title":"Runbox.Scenario.OutputAction.UpsertAssetAttributes","ref":"Runbox.Scenario.OutputAction.UpsertAssetAttributes.html"},{"type":"type","doc":"Attribute - access tag mapping.\n\nAccess tags are defined as map where keys are attributes and values\nare tags of these attributes.\n\nAttributes (keys) can be defined as:\n  * `:all` - all attributes in this upsert share the same access tags\n  * `String.t()` - single attribute access tags definition\n  * `[String.t()]` - list of attributes that share the same access tags\n\nAccess tags (values) can be defined by single access tag or list of access tags.\n\nUsing this map structure we can define shared access tag for multiple attributes\nas well as multiple access tag for single attribute (or combination of both).\n\nAttributes can be defined with `dot` syntax (e.g. `user.contact.address.street`)\nwhere nested attributes are separated by `.`. We can use this syntax to define\naccess tags for group of nested attributes - access tags defined for\n`user.contact.address` will be used for all nested attributes such as\n`user.contact.address.city` and `user.contact.address.street`.\n\nOverlapping access tags definitions are concatenated. For example access tags\ndefined as:\n\n```\n%UpsertAssetAttributes{\n  ...\n  access_tags: %{\n    all: [\"tag1\"],\n    \"user.contact\" => [\"tag2\"],\n    [\"user.contact.address\", \"user.contact.phone\"] => [\"tag3\"],\n    \"user.contact.address.street\" => \"tag4\"\n  }\n}\n```\n\nwould mean that:\n  * `\"user.contact\"` has these tags defined `[\"tag1\", \"tag2\"]`\n  * `\"user.contact.address.phone\"` has these tags defined `[\"tag1\", \"tag2\", \"tag3\"]`\n  * `\"user.contact.address.street\"` has these tags defined `[\"tag1\", \"tag2\", \"tag3\", \"tag4\"]`","title":"Runbox.Scenario.OutputAction.UpsertAssetAttributes.access_tags/0","ref":"Runbox.Scenario.OutputAction.UpsertAssetAttributes.html#t:access_tags/0"},{"type":"type","doc":"Upsert Asset Attributes\n\n  * `:type` - asset type (for more information, see `t:asset_type/0`)\n  * `:id` - asset ID (for more information, see `t:asset_id/0`)\n  * `:attributes` - a map of attributes in format `%{\"name\" => value}` to be\n    inserted or updated\n  * `:access_tags` - a map of attributes mapped to its access tags (for more information,\n    see `t:access_tags/0`)","title":"Runbox.Scenario.OutputAction.UpsertAssetAttributes.t/0","ref":"Runbox.Scenario.OutputAction.UpsertAssetAttributes.html#t:t/0"},{"type":"module","doc":"Parameters for output action Upsert Edge.\n\nThe resulting output action will create an edge between two assets. The\nassets do not have to exist at the time of the edge creation.","title":"Runbox.Scenario.OutputAction.UpsertEdge","ref":"Runbox.Scenario.OutputAction.UpsertEdge.html"},{"type":"type","doc":"Upsert Edge\n\n  * `:from_type` - type of the asset which this edge goes out from\n  * `:from_id` - id of the asset which this edge goes out from\n  * `:to_type` - type of the asset which this edge goes to\n  * `:to_id` - id of the asset which this edge goes to\n  * `:type` - type of the edge\n  * `:access_tags` - list of access tags for this edge","title":"Runbox.Scenario.OutputAction.UpsertEdge.t/0","ref":"Runbox.Scenario.OutputAction.UpsertEdge.html#t:t/0"},{"type":"behaviour","doc":"Sandbox is a helper for executing runs without started Altworx application.\n\n Sandbox can be used for scenarios unit testing and also for interactive development, e.g. in\n LiveBook.\n\n This is a scenario-type agnostic version of `Runbox.Runtime.Stage.Sandbox` and\n `Runbox.Runtime.Simple.Sandbox`. This module will identify the type of the scenario and will route\n the request to the appropriate Sandbox implementation.\n\n For interactive development of scenarios, run `iex --erl \"-runbox mode slave\" -S mix` from command\n line in your scenarios application root directory and then use Sandbox functions to execute your\n runs (of course use `recompile` before run execution if you changed scenario code).","title":"Runbox.Runtime.Sandbox","ref":"Runbox.Runtime.Sandbox.html"},{"type":"function","doc":"Starts a new run for `scenario_id` and processes all input messages.\n\nReturns all output actions generated by run execution. Input messages for each runtime topic,\nreferenced by its logical name, are passed in the `topics` argument. Undefined topics are\nconsidered empty.\n\nThere are few options that can be specified.\n\n- `:start_from` - specify the start from value for the run. The value is used to filter out\n  messages and is passed to the scenraio's `init` callback.\n- `:modules` - list of modules containing the scenario (manifest and template, which can be just\n  one module). If not specified, modules are auto-discovered from the current Mix app. This opens\n  up the possibility to write ad-hoc scenarios with only Runbox as dependency, e.g. in\n  [Livebook](https://livebook.dev/).","title":"Runbox.Runtime.Sandbox.execute_run/3","ref":"Runbox.Runtime.Sandbox.html#execute_run/3"},{"type":"callback","doc":"","title":"Runbox.Runtime.Sandbox.execute_run/3","ref":"Runbox.Runtime.Sandbox.html#c:execute_run/3"},{"type":"type","doc":"","title":"Runbox.Runtime.Sandbox.opt/0","ref":"Runbox.Runtime.Sandbox.html#t:opt/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Sandbox.opts/0","ref":"Runbox.Runtime.Sandbox.html#t:opts/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Sandbox.topic_name/0","ref":"Runbox.Runtime.Sandbox.html#t:topic_name/0"},{"type":"type","doc":"","title":"Runbox.Runtime.Sandbox.topics/0","ref":"Runbox.Runtime.Sandbox.html#t:topics/0"},{"type":"module","doc":"Sandbox is a helper for executing runs without started Altworx application.\n\nSandbox can be used for scenarios unit testing and also for interactive development, e.g. in\nLiveBook.\n\nThis module is only for Simple scenarios.\n\nFor interactive development of scenarios, run `iex --erl \"-runbox mode slave\" -S mix` from command\nline in your scenarios application root directory and then use Sandbox functions to execute your\nruns (of course use `recompile` before run execution if you changed scenario code).","title":"Runbox.Runtime.Simple.Sandbox","ref":"Runbox.Runtime.Simple.Sandbox.html"},{"type":"function","doc":"Starts a new run for `scenario_id` and processes all input messages.\n\nReturns all output actions generated by run execution. Input messages for each runtime topic,\nreferenced by its logical name, are passed in the `topics` argument. Undefined topics are\nconsidered empty.\n\nThere are few options that can be specified.\n\n- `:start_from` - specify the start from value for the run. The value is used to filter out\n  messages and is passed to the scenario's `init` callback.\n- `:modules` - list of modules containing the scenario (manifest and template, which can be just\n  one module). If not specified, modules are auto-discovered from the current Mix app. This opens\n  up the possibility to write ad-hoc scenarios with only Runbox as dependency, e.g. in\n  [Livebook](https://livebook.dev/).","title":"Runbox.Runtime.Simple.Sandbox.execute_run/3","ref":"Runbox.Runtime.Simple.Sandbox.html#execute_run/3"},{"type":"module","doc":"Sandbox is helper for executing runs without started Altworx application.\n\nSandbox can be used for scenarios unit testing and also for interactive\ndevelopment of scenarios. When used for unit testing, it can be used\ndirectly or via `Runbox.Runtime.Stage.Sandbox.TestRunner`.\n\nThis module is only for StageBased scenarios.\n\nFor interactive development of scenarios, run `iex --erl \"-runbox mode slave\" -S mix`\nfrom command line in your scenarios application root directory and then use Sandbox\nfunctions to execute your runs (of course use `recompile` before run execution if you\nchanged scenario code).","title":"Runbox.Runtime.Stage.Sandbox","ref":"Runbox.Runtime.Stage.Sandbox.html"},{"type":"function","doc":"Starts new run for `scenario_id` and waits until all input messages are processed.\n\nReturns all output actions generated by run execution. Input messages for each runtime\ntopic used by scenario should be defined in `topics` argument. Undefined topics are\nconsidered empty. Messages in each topic must be in ascending order by `timestamp`.\n\nThere are few options that can be specified.\n\n- `:start_from` - specify the start from value for the run. The value is used to filter out\n  messages (as in a real run, input topics only use messages with `timestamp >= start_from`) and\n  the value is provided to template's `init` callback.\n- `:modules` - list of modules containing scenarios (manifests, templates). If list of modules is\n  specified `execute_run` does not search scenario in `:scenarios` application release, but in\n  given modules. It opens possibility to write ad-hoc scenarios with only dependency on Runbox\n  application, for example use it in [Livebook](https://livebook.dev/). To see example of such\n  livebook just run `livebook server --home .` in runbox root directory, go to generated livebook\n  server link in browser and finally open `sandbox-scenario-demo.livemd` document. Running docker\n  or altworx is not needed, only runbox app (and its deps) is required!","title":"Runbox.Runtime.Stage.Sandbox.execute_run/3","ref":"Runbox.Runtime.Stage.Sandbox.html#execute_run/3"},{"type":"module","doc":"For new unit testing of scenarios it is recommended to use functions in\n`Runbox.Runtime.Stage.Sandbox` directly. This module should only help to\nconvert old scenarios integration tests (using Solutions.TestRunner) to unit\ntests. For examples of such conversion see tests in the scenarios master\nbranch.","title":"Runbox.Runtime.Stage.Sandbox.TestRunner","ref":"Runbox.Runtime.Stage.Sandbox.TestRunner.html"},{"type":"function","doc":"Runs an unit test according to the `test_def` definition.\n\nFunction accepts the same `test_def` definition as `Solutions.TestRunner.run_test/1`\nwith few differences:\n\n  * Only `run_options.scenario_id`, `run_options.start_from`, `messages` and `expected_actions`\n    properties are used, other properties are ignored.\n\n  * Instead of output log format `expected_actions` are `Runbox.Scenario.OutputAction`\n    structures, for example\n\n    ```\n    %{\n      action: :add_asset,\n      params: %{attributes: %{}, id: \"/diamonds/O\"},\n      result: \"nil\",\n      serialization_vector: [\"/diamonds/O\"],\n      side_effects: false,\n      timestamp: 1\n    }\n    ```\n\n    must be changed to:\n\n    ```\n    %Runbox.Scenario.OutputAction{\n      body: %Runbox.Scenario.OutputAction.UpsertAssetAttributes{\n        type: \"/diamonds\",\n        id: \"O\", attributes: %{}\n      },\n      timestamp: 1\n    }\n    ```\n\nTo see differences take a look at `deduplication_diamond_test.exs` in scenarios and\nthe same test in solutions.","title":"Runbox.Runtime.Stage.Sandbox.TestRunner.run_test/1","ref":"Runbox.Runtime.Stage.Sandbox.TestRunner.html#run_test/1"},{"type":"module","doc":"Set of utility functions for use in notification templates.","title":"Runbox.Notifications.TemplateHelper","ref":"Runbox.Notifications.TemplateHelper.html"},{"type":"function","doc":"Returns a URL to the asset detail.\n\nThe type of UI can be specified, and the link is generated to work with that UI. Can be `:new` or\n`:old`.","title":"Runbox.Notifications.TemplateHelper.asset_link/2","ref":"Runbox.Notifications.TemplateHelper.html#asset_link/2"},{"type":"function","doc":"Formats a number by separating thousands with a separator.","title":"Runbox.Notifications.TemplateHelper.format_number/2","ref":"Runbox.Notifications.TemplateHelper.html#format_number/2"},{"type":"function","doc":"Formats timestamp into the given format in the given timezone.","title":"Runbox.Notifications.TemplateHelper.format_ts/3","ref":"Runbox.Notifications.TemplateHelper.html#format_ts/3"},{"type":"macro","doc":"Wraps the given EEx content with a EEx template.\n\nTemplates can be bundled with a scenario or be a part of the system, `source`\nshould be set to `scenario_name` for scenario-based\ntemplates and `:system` for templates bundled with the system.\n\nThe templates that can be included, so called shared templates, are stored in\n`shared` folder\n(`scenarios/priv/notifications/{scenario_name}/shared`).\nEach shared template consists of two files - header and footer - which are\nwrapped around the given content. The filenames have `_header` and `_footer`\nappended, for `template_name = \"email\"` the two filenames are\n`email_header.eex` and `email_footer.eex`. System templates are stored in\n`apps/asset_map/templates/shared`.\n\nVia `context` you can pass additional parameters to the shared template\n(passed as assigns). The expected context depends on the template.","title":"Runbox.Notifications.TemplateHelper.include_template/4","ref":"Runbox.Notifications.TemplateHelper.html#include_template/4"},{"type":"macro","doc":"```\n<% alias Runbox.Notifications.TemplateHelper %>\n<% require TemplateHelper %>\n<%= TemplateHelper.include_template({\"test\", 1}, \"email_rich\", title: \"Test email\") do %>\n   Test Email \n   This text is surrounded by the template... \n<% end %>\n```","title":"Example usage - Runbox.Notifications.TemplateHelper.include_template/4","ref":"Runbox.Notifications.TemplateHelper.html#include_template/4-example-usage"},{"type":"function","doc":"Returns the UI url.","title":"Runbox.Notifications.TemplateHelper.ui_url/0","ref":"Runbox.Notifications.TemplateHelper.html#ui_url/0"},{"type":"module","doc":"Utilities for working with Asset","title":"Runbox.Scenario.Asset","ref":"Runbox.Scenario.Asset.html"},{"type":"function","doc":"Split asset id by the last slash.","title":"Runbox.Scenario.Asset.split_id/1","ref":"Runbox.Scenario.Asset.html#split_id/1"},{"type":"function","doc":"iex> Asset.split_id(\"/type/id\")\n    {\"/type\", \"id\"}\n    iex> Asset.split_id(\"id\")\n    {\"\", \"id\"}","title":"Example - Runbox.Scenario.Asset.split_id/1","ref":"Runbox.Scenario.Asset.html#split_id/1-example"},{"type":"module","doc":"A set of utility functions for Enums.","title":"Runbox.Utils.Enum","ref":"Runbox.Utils.Enum.html"},{"type":"function","doc":"Returns `:ok` if `fun` applied to every element of `enumerable` returns `:ok`.\nOtherwise it returns the first error occurred during `fun` call.","title":"Runbox.Utils.Enum.each_while_ok/2","ref":"Runbox.Utils.Enum.html#each_while_ok/2"},{"type":"function","doc":"iex> each_while_ok([], fn _ -> :not_called_at_all end)\n    :ok\n\n    iex> each_while_ok([1, 2, 3], fn _ -> :ok end)\n    :ok\n\n    iex> each_while_ok([1, 2, 3], &(if &1 == 2, do: {:error, &1}, else: :ok))\n    {:error, 2}","title":"Example - Runbox.Utils.Enum.each_while_ok/2","ref":"Runbox.Utils.Enum.html#each_while_ok/2-example"},{"type":"function","doc":"Returns `{:ok, list_of_values}` if `fun` applied to every element of\n`enumerable` returns `{:ok, value}`. Otherwise it returns the first\nerror occurred during `fun` call.","title":"Runbox.Utils.Enum.map_while_ok/2","ref":"Runbox.Utils.Enum.html#map_while_ok/2"},{"type":"function","doc":"iex> map_while_ok([], fn _ -> :not_called_at_all end)\n    {:ok, []}\n\n    iex> map_while_ok([1, 2, 3], &({:ok, &1 * 2}))\n    {:ok, [2, 4, 6]}\n\n    iex> map_while_ok([1, 2, 3], &(if &1 == 2, do: {:error, &1 * 2}, else: {:ok, &1 * 2}))\n    {:error, 4}","title":"Example - Runbox.Utils.Enum.map_while_ok/2","ref":"Runbox.Utils.Enum.html#map_while_ok/2-example"},{"type":"function","doc":"Invokes `fun` for each element in the `enumerable` with the accumulator.\nIf every `fun` call returns `{:ok, new_acc}`, functions returns `{:ok, last_acc}`.\nOtherwise it returns the first error occurred during `fun` call.","title":"Runbox.Utils.Enum.reduce_while_ok/3","ref":"Runbox.Utils.Enum.html#reduce_while_ok/3"},{"type":"function","doc":"iex> reduce_while_ok([], :initial, fn _, _ -> :not_called_at_all end)\n    {:ok, :initial}\n\n    iex> reduce_while_ok([1, 2, 3], 0, &({:ok, &1 + &2}))\n    {:ok, 6}\n\n    iex> reduce_while_ok([1, 2, 3], 0, &(if &1 == 2, do: {:error, &1 + &2}, else: {:ok, &1 + &2}))\n    {:error, 3}","title":"Example - Runbox.Utils.Enum.reduce_while_ok/3","ref":"Runbox.Utils.Enum.html#reduce_while_ok/3-example"},{"type":"module","doc":"A set of utility functions for Maps.","title":"Runbox.Utils.Map","ref":"Runbox.Utils.Map.html"},{"type":"function","doc":"Follows the `path` in the (deep) map, defaulting to `default` if some of the\nobjects on the path is missing.","title":"Runbox.Utils.Map.get_path/3","ref":"Runbox.Utils.Map.html#get_path/3"},{"type":"function","doc":"iex> object = %{\"some\" => %{\"deep\" => %{\"object\" => \"data\"}}}\n    ...> Runbox.Utils.Map.get_path(object, [\"some\", \"deep\", \"object\"])\n    \"data\"\n    iex> Runbox.Utils.Map.get_path(object, [\"non-existent\"], \"default\")\n    \"default\"","title":"Example - Runbox.Utils.Map.get_path/3","ref":"Runbox.Utils.Map.html#get_path/3-example"},{"type":"module","doc":"Utilities for handling configured paths.","title":"Runbox.Utils.Path","ref":"Runbox.Utils.Path.html"},{"type":"function","doc":"","title":"Runbox.Utils.Path.altworx_root_env_var/0","ref":"Runbox.Utils.Path.html#altworx_root_env_var/0"},{"type":"function","doc":"Gets the Altworx root dir.","title":"Runbox.Utils.Path.get_altworx_root/0","ref":"Runbox.Utils.Path.html#get_altworx_root/0"},{"type":"function","doc":"Resolves a possibly relative path to absolute path by prepending the Altworx\nroot dir (see `set_altworx_root/1`).","title":"Runbox.Utils.Path.resolve_path/1","ref":"Runbox.Utils.Path.html#resolve_path/1"},{"type":"function","doc":"iex> Path.set_altworx_root(\"/opt/altworx\")\n    :ok\n    iex> Path.resolve_path(\"/already/absolute/path\")\n    \"/already/absolute/path\"\n    iex> Path.resolve_path(\"some/../relative/path\")\n    \"/opt/altworx/relative/path\"","title":"Examples - Runbox.Utils.Path.resolve_path/1","ref":"Runbox.Utils.Path.html#resolve_path/1-examples"},{"type":"function","doc":"Sets where the Altworx root dir is. Configuration and data dirs are supposed\nto be relative to the root dir. It is typically the directory which the VM is\nstarted in.","title":"Runbox.Utils.Path.set_altworx_root/1","ref":"Runbox.Utils.Path.html#set_altworx_root/1"},{"type":"module","doc":"","title":"Runbox.Utils.Traversal","ref":"Runbox.Utils.Traversal.html"},{"type":"function","doc":"Performs a depth-first pre-order traversal on the given `value`.\n\nTransforms each subterm via the given `fun`, before descending into its descendants.","title":"Runbox.Utils.Traversal.prewalk/2","ref":"Runbox.Utils.Traversal.html#prewalk/2"},{"type":"function","doc":"iex> Traversal.prewalk(\n    ...>    [1, 2, :pi, 4, %{5 => [6, 7]}],\n    ...>    fn\n    ...>      x when is_integer(x) -> Integer.to_string(x)\n    ...>      x when is_list(x) -> Enum.reverse(x)\n    ...>      x -> x\n    ...>    end\n    ...> )\n    [%{\"5\" => [\"7\", \"6\"]}, \"4\", :pi, \"2\", \"1\"]","title":"Example - Runbox.Utils.Traversal.prewalk/2","ref":"Runbox.Utils.Traversal.html#prewalk/2-example"},{"type":"exception","doc":"","title":"Runbox.Runtime.Stage.Emulator.EmulationError","ref":"Runbox.Runtime.Stage.Emulator.EmulationError.html"},{"type":"extras","doc":"# Altworx Runbox\n\n**Runbox** is a runtime environment for Altworx scenarios.\nIt provides essential data structures, APIs, and tools for developing, testing, and packaging scenarios.\n\nCheck out also [Toolbox](https://hex.pm/packages/altworx_toolbox), which builds on top of Runbox and\nprovides conveniences and additional abstractions for writing scenarios.\n\nNote that there are two types of documentation for Runbox.\nThe [main documentation](https://hexdocs.pm/altworx_runbox) (the one you are probably seeing now)\ncovers all user-facing modules. This is the go-to documentation when writing scenarios.\nHowever, there is also the [internal documentation](https://hexdocs.pm/altworx_runbox/internal)\nthat covers also internal modules. That documentation might be useful if you want to dig deeper\ninto how scenarios run.","title":"Altworx Runbox","ref":"readme.html"},{"type":"extras","doc":"Scenarios in Altworx are developed as Elixir/Mix projects, with a single project potentially containing\nmultiple scenarios. When packaged, all scenarios in a project are bundled together.\n\nTo start a new scenario project, create a blank project using `mix new project_name`, then add Runbox\nas a dependency in your `mix.exs`:\n\n```elixir\ndefp deps do\n  [\n    {:runbox, \"~> 9.0\", hex: :altworx_runbox}\n  ]\nend\n```","title":"Scenario projects - Altworx Runbox","ref":"readme.html#scenario-projects"},{"type":"extras","doc":"Scenario is a collection of Elixir modules. The main module is a manifest module. Manifest module\ndescribes the scenario - carries its metadata. Manifest module can be named whatever you like but \nit needs to explicitly implement the `Runbox.Scenario.Manifest` behaviour with one limitation. Two \nnames of manifest modules can not be prefixes of each other (e.g. `Some.Scenario` and \n`Some.Scenario.Specified`) as it might interfere with template detection of these scenarios. If such\ncase is detected, both respective scenarios will not be loaded.\n\nThen a scenario needs modules which define it's behaviour - what the scenario does. These modules\nand their structure depend on the scenario type. There may be multiple different scenario types.\nEach can behave and run differently and every type has it's own strengths and weaknesses. Available\ntypes are defined in `Runbox.Scenario.Type`.\n\nFor more in-depth details about how scenario works and how to write one, see the following modules.\n- `Simple` scenario runtime - see `Runbox.Scenario.Simple`\n- `StageBased` scenario runtime - see `Runbox.Scenario.Template.StageBased`\n\nScenario can define and use other modules. These modules don't have any predefined meaning nor do we\nrequire any specific module names.","title":"Scenario structure - Altworx Runbox","ref":"readme.html#scenario-structure"},{"type":"extras","doc":"Runbox requires the `:scenario_config_dir` environment variable under the :runbox app to be set.\nThis is automatically handled when running in Altworx but may need to be configured for the `test`\nenvironment. You can use a dummy value for testing, like so in `config/test.exs`:\n\n```elixir\nimport Config\n\nconfig :runbox, scenario_config_dir: \"/non-existing/does/not/matter\"\n```","title":"Configuration - Altworx Runbox","ref":"readme.html#configuration"},{"type":"extras","doc":"Using Runbox, scenarios are packaged into *artefacts* (also called *releases*) that can be deployed\nto and run in Altworx. These artefacts include both the scenarios code and the runtime environment.\nTo build a scenario artefact, make sure you have Docker installed and then run\n\n```sh\nmix altworx.scenario_release\n```\n\nYou can then load the built scenario artifact in the Admin UI. Do not forget to restart Altworx\nafterwards.","title":"Building scenario artefacts - Altworx Runbox","ref":"readme.html#building-scenario-artefacts"},{"type":"extras","doc":"Altworx runs in a single BEAM node and spawns a separate BEAM node for each scenario release.\nBoth Altworx and the scenario nodes run the Runbox app, so Runbox also serves for communication\nbetween these two. On the Altworx node, it is responsible for starting the scenario nodes.\n\nScenario nodes can are used for:\n* listing scenarios in a scenario release\n* spawning components for a scenario \"run\"\n* processing notification templates","title":"Architectural details - Altworx Runbox","ref":"readme.html#architectural-details"},{"type":"extras","doc":"# Changelog","title":"Changelog","ref":"changelog.html"},{"type":"extras","doc":"","title":"Master - Changelog","ref":"changelog.html#master"},{"type":"extras","doc":"","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"16.0.0 - Changelog","ref":"changelog.html#16-0-0"},{"type":"extras","doc":"- scenario: Let user specify emails to reply to\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/50),\n   [#7024](https://altworx.plan.io/issues/7024))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"15.0.0 - Changelog","ref":"changelog.html#15-0-0"},{"type":"extras","doc":"- Add access_tags to Scenario.OutputActions\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/49),\n   [#6937](https://altworx.plan.io/issues/6937))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"14.1.0 - Changelog","ref":"changelog.html#14-1-0"},{"type":"extras","doc":"- Naming convention for manifest modules are no longer mandatory but the \n  module now needs to explicitly implement the `Runbox.Scenario.Manifest`\n  behavior (`use Runbox.Scenario.Manifest`) to be recognized as valid","title":"Warnings - Changelog","ref":"changelog.html#warnings"},{"type":"extras","doc":"- Remove naming requirementes for manifest modules\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/48),\n   [#6750](https://altworx.plan.io/issues/6750))\n\n- Generate id using UUIDv4 instead of UUIDv1\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/47),\n   [#6847](https://altworx.plan.io/issues/6847))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"14.0.1 - Changelog","ref":"changelog.html#14-0-1"},{"type":"extras","doc":"- Unplanned Incident future can specify nil timestamp.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/46),\n   [#6866](https://altworx.plan.io/issues/6866))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"14.0.0 - Changelog","ref":"changelog.html#14-0-0"},{"type":"extras","doc":"- Function `Runbox.Runtime.OutputAction.is_oa_body?` was renamed to\n  `Runbox.Runtime.OutputAction.oa_body?`.","title":"Warnings - Changelog","ref":"changelog.html#warnings"},{"type":"extras","doc":"- Upgrade Elixir to 1.18.1 and OTP to 27.2\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/45),\n   [#6782](https://altworx.plan.io/issues/6782))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"13.0.3 - Changelog","ref":"changelog.html#13-0-3"},{"type":"extras","doc":"- Fix logger configuration for Slave VMs\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/44),\n   [#6789](https://altworx.plan.io/issues/6789))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"13.0.2 - Changelog","ref":"changelog.html#13-0-2"},{"type":"extras","doc":"- add partition to Runbox.Message origin\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/43),\n   [#6832](https://altworx.plan.io/issues/6832))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"13.0.1 - Changelog","ref":"changelog.html#13-0-1"},{"type":"extras","doc":"- Add Sandbox for Simple runtime and generic Sandbox module\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/42),\n   [#6774](https://altworx.plan.io/issues/6774))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"13.0.0 - Changelog","ref":"changelog.html#13-0-0"},{"type":"extras","doc":"- The Toolbox structs have been finally renamed to their proper names. You have to rename them\n  in your scenario code, normalizers, and `config.ini` as well. Occurrences in scenario savepoints\n  will be renamed automatically. The list of the renamed structs is the following:\n    - `Toolbox.Message` → `Runbox.Message` (used in scenarios and normalizers)\n    - `Toolbox.Runtime.Stage.Unit` → `Runbox.Runtime.Stage.Unit` (used in StageBased scenarios)\n    - `Toolbox.Scenario.UserAction` → `Runbox.Scenario.UserAction` (used in scenarios and\n      pipeline definition inside `config.ini`)\n    - `Toolbox.JitterMessage` → `Runbox.JitterMessage` (probably not used in scenario code)\n    - `Toolbox.ErrorMessage` → `Runbox.ErrorMessage` (probably not used in scenario code)","title":"Warnings - Changelog","ref":"changelog.html#warnings"},{"type":"extras","doc":"- Move Toolbox structs to the Runbox namespace (see warnings above)\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/41),\n   [#5849](https://altworx.plan.io/issues/5849), [#5850](https://altworx.plan.io/issues/5850),\n   [#5851](https://altworx.plan.io/issues/5851), [#5852](https://altworx.plan.io/issues/5852),\n   [#5853](https://altworx.plan.io/issues/5853))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"12.1.0 - Changelog","ref":"changelog.html#12-1-0"},{"type":"extras","doc":"- `Runbox.Runtime.Simple.StageBasedEmulator` requires Altworx 24.11+.","title":"Warnings - Changelog","ref":"changelog.html#warnings"},{"type":"extras","doc":"- Fix timeout execution order in Simple scenario\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/40))\n- Make Timezip more deterministic: messages in the same timestamp are produced in the topology order\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/40))\n- Add StageBased scenario single-process emulator\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/40),\n   [#6759](https://altworx.plan.io/issues/6759))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"12.0.0 - Changelog","ref":"changelog.html#12-0-0"},{"type":"extras","doc":"- The `unit_id` field of the `Runbox.Runtime.RuntimeInstruction.Timeout` struct\n  was dropped, as runtime was just ignoring it.","title":"Warnings - Changelog","ref":"changelog.html#warnings"},{"type":"extras","doc":"- Add support for timeouts in the Simple scenario runtime\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/38),\n   [#6695](https://altworx.plan.io/issues/6695))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"11.0.1 - Changelog","ref":"changelog.html#11-0-1"},{"type":"extras","doc":"- Decrease timeout for starting peer nodes to 15 seconds\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/39))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"11.0.0 - Changelog","ref":"changelog.html#11-0-0"},{"type":"extras","doc":"- Add documentation for asset types and IDs\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/37))\n- Add Simple scenario runtime\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/35),\n   [#6694](https://altworx.plan.io/issues/6694))\n- Add incident actors to Event Output Action\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/36),\n   [#6591](https://altworx.plan.io/issues/6591))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"10.0.0 - Changelog","ref":"changelog.html#10-0-0"},{"type":"extras","doc":"- Update the Livebook example\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/30))\n- Require unit ID to be explicitly set when creating the Unit struct\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/31),\n   [#6609](https://altworx.plan.io/issues/6609))\n- Improve the documentation\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/34),\n   [#6195](https://altworx.plan.io/issues/6195))\n- Use peer instead of deprecated slave module for scenario's nodes\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/33),\n   [#5918](https://altworx.plan.io/issues/5918))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"9.0.0 - Changelog","ref":"changelog.html#9-0-0"},{"type":"extras","doc":"- Notifications specifications are no longer retrieved from `spec.exs` file but\n  are now part of Manifest, all scenarios should now define their notifications\n  specifications directly in scenario manifest.","title":"Warnings - Changelog","ref":"changelog.html#warnings"},{"type":"extras","doc":"- Improve docs for `Runbox.Scenario.OutputAction.Event`\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/26))\n- Move notification specification to scenario manifest\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/28),\n   [#6569](https://altworx.plan.io/issues/6569))\n- Better separate scenario runtimes\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/29),\n   [#6665](https://altworx.plan.io/issues/6665))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"8.0.0 - Changelog","ref":"changelog.html#8-0-0"},{"type":"extras","doc":"- Add additional event parameters to Incident History OA\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/27),\n   [#6592](https://altworx.plan.io/issues/6592))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"7.0.1 - Changelog","ref":"changelog.html#7-0-1"},{"type":"extras","doc":"- Anticipate errors from scenario manifest\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/23),\n   [#6312](https://altworx.plan.io/issues/6312))\n- Explicitly propagate altworx root dir to slave\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/24),\n   [#6331](https://altworx.plan.io/issues/6331))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"7.0.0 - Changelog","ref":"changelog.html#7-0-0"},{"type":"extras","doc":"- Scenarios are now built with Elixir 1.15.7 and OTP 26.2.1. There may be some breaking\n  changes for your code (see https://hexdocs.pm/elixir/1.15/changelog.html and\n  https://www.erlang.org/downloads/26). It is recommended to update your `.tool-versions`\n  to match these versions.","title":"Warnings - Changelog","ref":"changelog.html#warnings"},{"type":"extras","doc":"- Forbid slashes in IDs contained in asset/incident creation OAs\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/22),\n   [#6327](https://altworx.plan.io/issues/6327))\n- Build scenarios with Elixir 1.15 and OTP 26 (to be compatible with Altworx 24.3)\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/21),\n   [#6285](https://altworx.plan.io/issues/6285))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"6.0.0 - Changelog","ref":"changelog.html#6-0-0"},{"type":"extras","doc":"- All output actions have changed. The following snippet shows how should be\n  the output actions updated from the old format to the new one.\n  ```elixir\n  alias Runbox.Scenario.OutputAction, as: OA\n\n  # Create Asset\n  # old\n  OA.new_create_asset_action(\"/asset/camera/1\", 1, %{\"a\" => 1})\n  # new\n  %OA.UpsertAssetAttributes{type: \"/asset/camera\", id: \"1\", attributes: %{\"a\" => 1}}\n\n  # Update Asset\n  # old - deletes and updates attributes via a single action\n  OA.new_update_asset_action(\"/asset/camera/1\", 1, %{\"a\" => 1}, %{\"b\" => true})\n  # new - one action for delete, another one for upsert\n  %OA.DeleteAssetAttributes{type: \"/asset/camera\", id: \"1\", attributes: %{\"b\" => true}}\n  %OA.UpsertAssetAttributes{type: \"/asset/camera\", id: \"1\", attributes: %{\"a\" => 1}}\n\n  # Upsert Asset\n  # old\n  OA.new_upsert_asset_action(\"/asset/camera/1\", 1, %{\"a\" => 1})\n  # new\n  %OA.UpsertAssetAttributes{type: \"/asset/camera\", id: \"1\", attributes: %{\"a\" => 1}}\n\n  # Delete Asset\n  # old\n  OA.new_delete_asset_action(\"/asset/camera/1\", 1)\n  # new\n  %OA.DeleteAllAssetAttributes{type: \"/asset/camera\", id: \"1\"}\n\n  # Create Edge\n  # old\n  OA.new_create_edge_action(1, \"/asset/camera/1\", \"/asset/room/A\", \"observes\")\n  # new\n  %OA.UpsertEdge{\n    from_type: \"/asset/camera\",\n    from_id: \"1\",\n    to_type: \"/asset/room\",\n    to_id: \"A\",\n    type: \"observes\"\n  }\n\n  # Upsert Edge\n  # old\n  OA.new_upsert_edge_action(1, \"/asset/camera/1\", \"/asset/room/A\", \"observes\")\n  # new\n  %OA.UpsertEdge{\n    from_type: \"/asset/camera\",\n    from_id: \"1\",\n    to_type: \"/asset/room\",\n    to_id: \"A\",\n    type: \"observes\"\n  }\n\n  # Delete Edge\n  # old\n  OA.new_delete_edge_action(1, \"/asset/camera/1\", \"/asset/room/A\", \"observes\")\n  # new\n  %OA.DeleteEdge{\n    from_type: \"/asset/camera\",\n    from_id: \"1\",\n    to_type: \"/asset/room\",\n    to_id: \"A\",\n    type: \"observes\"\n  }\n\n  # Notification\n  # old\n  OA.new_notification_action(1, %{type: \"created_incident\", data: %{\"a\" => 1}})\n  # new\n  %OA.Notification{type: \"created_incident\", data: %{\"a\" => 1}}\n\n  # Event\n  # old\n  OA.new_event_action(1, %{\n    type: \"event_type\",\n    template: \"${actors.wagon}\",\n    actors: %{\"wagon\" => %{asset_type: \"/assets/wagon\", asset_id: \"123\"}}\n  })\n  # new\n  %OA.Event{\n    type: \"event_type\",\n    template: \"${actors.wagon}\",\n    actors: %{\"wagon\" => %{asset_type: \"/assets/wagon\", asset_id: \"123\"}}\n  }\n\n  # Execute SQL\n  # old\n  OA.execute_sql_action(1, %{\n    db_connection: db_connection,\n    sql_query: \"SELECT 1\",\n    data: [],\n    type: :postgresql\n  })\n  # new\n  %OA.ExecuteSQL{\n    db_connection: db_connection,\n    sql_query: \"SELECT 1\",\n    data: [],\n    type: :postgresql\n  }\n\n  # Create Incident\n  # old\n  OA.new_create_incident_action(1, %OA.Incident{\n    type: \"type\",\n    id: \"id\",\n    subject: \"subject\",\n    status: \"status\",\n    resolved: false,\n    severity: 1,\n    future: [],\n    history: [%OA.IncidentHistory{\n      status: \"status\",\n      severity: 1,\n      timestamp: 1,\n      description: \"description\"\n    }]\n  })\n  # new\n  %OA.Incident{\n    type: \"type\",\n    id: \"id\",\n    subject: \"subject\",\n    status: \"status\",\n    resolved: false,\n    severity: 1,\n    future: [],\n    history: [%OA.IncidentHistory{\n      status: \"status\",\n      severity: 1,\n      timestamp: 1,\n      description: \"description\"\n    }]\n  }\n\n  # Update Incident\n  # old\n  OA.new_update_incident_action(1, %OA.IncidentPatch{\n    type: \"type,\n    id: \"id\",\n    severity: 2\n  })\n  # new\n  %OA.IncidentPatch{\n    type: \"type,\n    id: \"id\",\n    severity: 2\n  }\n  ```\n\n  Refer to the documentation in `Runbox.Scenario.OutputAction` for details.\n\n  There is also a helper `Runbox.Scenario.Asset.split_id/1`. It splits the full\n  asset ID to asset type and asset ID.\n\n- The sandbox tests have changed the format of the produced output action.\n  Previously it was a struct\n\n  ```elixir\n  %Runbox.Scenario.OutputAction{\n    type: type,\n    timestamp: ts,\n    body: body\n  }\n  ```\n\n  Now the `:type` is gone and `:body` always contains a struct with the output\n  action parameters, e.g.\n\n  ```elixir\n  %Runbox.Scenario.OutputAction{\n    timestamp: ts,\n    body: %Runbox.Scenario.OutputAction.UpsertAssetAttributes{type: type, id: id, attributes: attrs}\n  }\n  ```","title":"Breaking Changes - Changelog","ref":"changelog.html#breaking-changes"},{"type":"extras","doc":"- New output actions for asset and edge.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/19),\n   [#6129](https://altworx.plan.io/issues/6129))\n- Rework of output action creation and handling.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/20),\n   [#6177](https://altworx.plan.io/issues/6177))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"5.0.0 - Changelog","ref":"changelog.html#5-0-0"},{"type":"extras","doc":"- Remove deprecation for '/incident' type.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/18),\n   [#5999](https://altworx.plan.io/issues/6066))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"- Removed code which separated \"/incident\" prefix from params's type","title":"Breaking changes - Changelog","ref":"changelog.html#breaking-changes"},{"type":"extras","doc":"","title":"4.0.0 - Changelog","ref":"changelog.html#4-0-0"},{"type":"extras","doc":"- New output actions for incidents. Also remove/update obsolete code.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/17),\n   [#5999](https://altworx.plan.io/issues/5999))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"- Removed code, which is likely not used anywhere, but check the scenarios.\n  - `Runbox.Scenario.OutputAction.new_incident_action/2`\n  - `%Runbox.Scenario.OutputAction.Incident{}`\n  - `%Runbox.Scenario.OutputAction.IncidentState{}`\n\n- New output actions of types `:create_incident` and `:update_incident`.\n  Altworx has to know how to handle those when depending on this version.","title":"Breaking changes - Changelog","ref":"changelog.html#breaking-changes"},{"type":"extras","doc":"","title":"3.0.0 - Changelog","ref":"changelog.html#3-0-0"},{"type":"extras","doc":"- Move certain modules from Toolbox to Runbox\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/16),\n   [#5848](https://altworx.plan.io/issues/5848))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"- Many modules were moved from Toolbox to Runbox (some kept their name, but the majority\n  of them was also renamed to be under the Runbox \"namespace\"), here are the ones that\n  affect scenario authors (for a complete list, see the task):\n  - Modules that will require changes in aliases (just replace `Toolbox` with `Runbox`)\n    - `Toolbox.Notifications.TemplateHelper` → `Runbox.Notifications.TemplateHelper`\n    - `Toolbox.Runtime.RuntimeInstruction` → `Runbox.Runtime.RuntimeInstruction`\n    - `Toolbox.Utils.Path` → `Runbox.Utils.Path`\n    - `Toolbox.Scenario.Config` → `Runbox.Scenario.Config`\n    - `Toolbox.Scenario.Helper` → `Runbox.Scenario.Helper`\n    - `Toolbox.Scenario.Manifest` → `Runbox.Scenario.Manifest`\n  - A module that requires change of `@behaviour` (just replace `Toolbox` with `Runbox`)\n    - `Toolbox.Scenario.Template.StageBased`→ `Runbox.Scenario.Template.StageBased`\n  - Configuration key move\n    - `:scenario_config_dir` config key was moved from `:toolbox` to `:runbox`,\n      you may have to change this in your `config/test.exs`","title":"Breaking changes - Changelog","ref":"changelog.html#breaking-changes"},{"type":"extras","doc":"","title":"2.1.0 - Changelog","ref":"changelog.html#2-1-0"},{"type":"extras","doc":"- Remove runbox version check\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/15),\n   [#5828](https://altworx.plan.io/issues/5828))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"2.0.0 - Changelog","ref":"changelog.html#2-0-0"},{"type":"extras","doc":"- Build scenarios with Elixir 1.14 and OTP 25 (to be compatible with Altworx 23.3+)\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/14),\n   [#5799](https://altworx.plan.io/issues/5907))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"1.4.1 - Changelog","ref":"changelog.html#1-4-1"},{"type":"extras","doc":"- Bump toolbox dependency to 2.0.0","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"1.4.0 - Changelog","ref":"changelog.html#1-4-0"},{"type":"extras","doc":"- Start slave VM as a hidden node\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/11),\n   [#5799](https://altworx.plan.io/issues/5799))\n- Check that scenario timeouts are not registered to the past.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/10),\n   [#5827](https://altworx.plan.io/issues/5827))\n- Use timeout message timestamp to register scenario timeouts.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/10),\n   [#5827](https://altworx.plan.io/issues/5827))\n- Check input messages order prior sandbox run execution.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/10),\n   [#5827](https://altworx.plan.io/issues/5827))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"- API function to register timeouts `Toolbox.Runtime.RuntimeInstruction.register_timeout/2`\n  was removed, update scenarios to use `Toolbox.Runtime.RuntimeInstruction.register_timeout/1`\n  instead.\n- Field `timeout` was removed from `Toolbox.Runtime.RuntimeInstruction.Timeout`","title":"Breaking changes - Changelog","ref":"changelog.html#breaking-changes"},{"type":"extras","doc":"","title":"1.3.0 - Changelog","ref":"changelog.html#1-3-0"},{"type":"extras","doc":"- Remove `Runbox.Scenario.Template` and use only `Runbox.ScenarioTemplate`.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/9))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"1.2.0 - Changelog","ref":"changelog.html#1-2-0"},{"type":"extras","doc":"- Default mode is `:slave`. Scenario repositories do not need to set this\n  anymore. Altworx needs to set the `:master` mode explicitly.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/8),\n   [#5714](https://altworx.plan.io/issues/5714))\n- Find scenario modules for sandbox testing from the current mix app. Scenario\n  repositories do not need to set `:scenario_app` app env anymore even for\n  testing.\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/8),\n   [#5714](https://altworx.plan.io/issues/5714))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"1.1.0 - Changelog","ref":"changelog.html#1-1-0"},{"type":"extras","doc":"- Mix task for building scenario releases\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/5),\n   [#5713](https://altworx.plan.io/issues/5713))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"- There is no need to set `:runbox, scenario_app: :app_name` in the scenario app\n  anymore. New task `mix altworx.scenario_release` does it automatically.\n\n- There is no need to set `releases` key in the `project/1` of `mix.exs` in the\n  scenario app. In fact, the `release` key is ignored when using the new mix\n  task `mix altworx.scenario_release`.","title":"Additional info - Changelog","ref":"changelog.html#additional-info"},{"type":"extras","doc":"","title":"1.0.0 - Changelog","ref":"changelog.html#1-0-0"},{"type":"extras","doc":"- Relax runbox version check\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/7),\n   [#5817](https://altworx.plan.io/issues/5817))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"0.1.3 - Changelog","ref":"changelog.html#0-1-3"},{"type":"extras","doc":"- Add support for start_from in Sandbox\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/4),\n   [#5821](https://altworx.plan.io/issues/5821))\n- Explicitly load modules when checking if a function is exported\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/6)\n   [#5784](https://altworx.plan.io/issues/5784))\n- Run slaves in interactive mode\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/6)\n   [#5784](https://altworx.plan.io/issues/5784))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"- Bump toolbox dependency to 1.0.0","title":"0.1.2 - Changelog","ref":"changelog.html#0-1-2"},{"type":"extras","doc":"","title":"0.1.1 - Changelog","ref":"changelog.html#0-1-1"},{"type":"extras","doc":"- Use runbox from hex in demo livebook\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/2))\n- Fix load topic in Sandbox\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/3))","title":"Changes - Changelog","ref":"changelog.html#changes"},{"type":"extras","doc":"","title":"0.1.0 - Changelog","ref":"changelog.html#0-1-0"},{"type":"extras","doc":"- Runbox is in separate repository\n  ([PR](https://gitlab.dev.altworx.com/erlang/product/runbox/-/merge_requests/1),\n   [PR](https://gitlab.dev.altworx.com/erlang/product/altworx/-/merge_requests/1853),\n   [#5711](https://altworx.plan.io/issues/5711))","title":"Changes - Changelog","ref":"changelog.html#changes"}],"producer":{"name":"ex_doc","version":[48,46,51,52,46,50]},"content_type":"text/markdown"}